/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "quakedef.h"
#include "r_local.h"
#ifdef PROTO
#define MAX_PARTICLES			8192	// overloaded
#else
#define MAX_PARTICLES			8192	// default max # of particles at one
										//  time
#endif
#define ABSOLUTE_MIN_PARTICLES	512		// no fewer than this no matter what's
										//  on the command line

int		ramp1[8] = {0x6f, 0x6d, 0x6b, 0x69, 0x67, 0x65, 0x63, 0x61};
int		ramp2[8] = {0x6f, 0x6e, 0x6d, 0x6c, 0x6b, 0x6a, 0x68, 0x66};
int		ramp3[8] = {0x6d, 0x6b, 6, 5, 4, 3};

particle_t	*active_particles, *free_particles;

particle_t	*particles;
int			r_numparticles;

vec3_t			r_pright, r_pup, r_ppn;

extern	int	particlespray;
extern	int	particleblood;
extern	int	particleset;
extern	int	particlelit;

/*
===============
R_InitParticles
===============
*/
void R_InitParticles (void)
{
	int		i;

	i = COM_CheckParm ("-particles");

	if (i)
	{
		r_numparticles = (int)(Q_atoi(com_argv[i+1]));
		if (r_numparticles < ABSOLUTE_MIN_PARTICLES)
			r_numparticles = ABSOLUTE_MIN_PARTICLES;
	}
	else
	{
		r_numparticles = MAX_PARTICLES;
	}

	particles = (particle_t *)
			Hunk_AllocName (r_numparticles * sizeof(particle_t), "particles");

}

#ifdef QUAKE2
void R_DarkFieldParticles (entity_t *ent)
{
	int			i, j, k;
	particle_t	*p;
	float		vel;
	vec3_t		dir;
	vec3_t		org;

	org[0] = ent->origin[0];
	org[1] = ent->origin[1];
	org[2] = ent->origin[2];
	for (i=-16 ; i<16 ; i+=8)
		for (j=-16 ; j<16 ; j+=8)
			for (k=0 ; k<32 ; k+=8)
			{
				if (!free_particles)
					return;
				p = free_particles;
				free_particles = p->next;
				p->next = active_particles;
				active_particles = p;

				p->die = cl.time + 0.2 + (rand()&7) * 0.02;
				p->color = 150 + rand()%6;
				p->type = pt_slowgrav;

				dir[0] = j*8;
				dir[1] = i*8;
				dir[2] = k*8;

				p->org[0] = org[0] + i + (rand()&3);
				p->org[1] = org[1] + j + (rand()&3);
				p->org[2] = org[2] + k + (rand()&3);

				VectorNormalize (dir);
				vel = 50 + (rand()&63);
				VectorScale (dir, vel, p->vel);
			}
}
#endif


/*
===============
R_EntityParticles
===============
*/

#define NUMVERTEXNORMALS	162
extern	float	r_avertexnormals[NUMVERTEXNORMALS][3];
vec3_t	avelocities[NUMVERTEXNORMALS];
float	beamlength = 16;
vec3_t	avelocity = {23, 7, 3};
float	partstep = 0.01;
float	timescale = 0.01;

void R_EntityParticles (entity_t *ent)
{
//	int			count;	// 2001-12-10 Reduced compiler warnings by Jeff Ford
	int			i;
	particle_t	*p;
	float		angle;
	float		sp, sy, cp, cy;	//sr, cr, 	// 2001-12-10 Reduced compiler warnings by Jeff Ford
	vec3_t		forward;
	float		dist;

	dist = 64;
//	count = 50;	// 2001-12-10 Reduced compiler warnings by Jeff Ford

	if (!avelocities[0][0])
	{
		for (i=0 ; i<NUMVERTEXNORMALS*3 ; i++)
			avelocities[0][i] = (rand()&255) * 0.01;
	}

	for (i=0 ; i<NUMVERTEXNORMALS ; i++)
	{
		angle = cl.time * avelocities[i][0];
		sy = sin(angle);
		cy = cos(angle);
		angle = cl.time * avelocities[i][1];
		sp = sin(angle);
		cp = cos(angle);
		angle = cl.time * avelocities[i][2];
//		sr = sin(angle);	// 2001-12-10 Reduced compiler warnings by Jeff Ford
//		cr = cos(angle);	// 2001-12-10 Reduced compiler warnings by Jeff Ford

		forward[0] = cp*cy;
		forward[1] = cp*sy;
		forward[2] = -sp;

		if (!free_particles)
			return;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 0.01;
		p->color = 0x6f;
		p->type = pt_explode;

		p->org[0] = ent->origin[0] + r_avertexnormals[i][0]*dist + forward[0]*beamlength;
		p->org[1] = ent->origin[1] + r_avertexnormals[i][1]*dist + forward[1]*beamlength;
		p->org[2] = ent->origin[2] + r_avertexnormals[i][2]*dist + forward[2]*beamlength;
	}
}


/*
===============
R_ClearParticles
===============
*/
void R_ClearParticles (void)
{
	int		i;

	free_particles = &particles[0];
	active_particles = NULL;

	for (i=0 ;i<r_numparticles ; i++)
		particles[i].next = &particles[i+1];
	particles[r_numparticles-1].next = NULL;
}


void R_ReadPointFile_f (void)
{
	FILE	*f;
	vec3_t	org;
	int		r;
	int		c;
	particle_t	*p;
	char	name[MAX_OSPATH];

	sprintf (name,"maps/%s.pts", sv.name);

	COM_FOpenFile (name, &f, NULL);	// 2001-09-12 Returning from which searchpath a file was loaded by Maddes
	if (!f)
	{
		Con_Printf ("couldn't open %s\n", name);
		return;
	}

	Con_Printf ("Reading %s...\n", name);
	c = 0;
	for ( ;; )
	{
		r = fscanf (f,"%f %f %f\n", &org[0], &org[1], &org[2]);
		if (r != 3)
			break;
		c++;

		if (!free_particles)
		{
			Con_Printf ("Not enough free particles\n");
			break;
		}
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = 99999;
		p->color = (-c)&15;
		p->type = pt_static;
		VectorCopy (vec3_origin, p->vel);
		VectorCopy (org, p->org);
	}

	fclose (f);
	Con_Printf ("%i points read\n", c);
}

/*
===============
R_ParseParticleEffect

Parse an effect out of the server message
===============
*/
void R_ParseParticleEffect (void)
{
	vec3_t		org, dir;
	int			i, count, msgcount, color;

	for (i=0 ; i<3 ; i++)
		org[i] = MSG_ReadCoord ();
	for (i=0 ; i<3 ; i++)
		dir[i] = MSG_ReadChar () * (1.0/16);
	msgcount = MSG_ReadByte ();
	color = MSG_ReadByte ();

if (msgcount == 255)
	count = 1024;
else
	count = msgcount;

if (particlespray)
	R_RunWarticleEffect (org, dir, color, count);
else
	R_RunParticleEffect (org, dir, color, count);
}

/*
===============
R_ParticleExplosion

===============
*/
void R_ParticleExplosion (vec3_t org)
{
	int			i, j;
	particle_t	*p;

	for (i=0 ; i<1024 ; i++)
	{
		if (!free_particles)
			return;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 5;
		p->color = ramp1[0];
		p->ramp = rand()&3;
		if (i & 1)
		{
			p->type = pt_explode;
			for (j=0 ; j<3 ; j++)
			{
				p->org[j] = org[j] + ((rand()%32)-16);
				p->vel[j] = (rand()%512)-256;
			}
		}
		else
		{
			p->type = pt_explode2;
			for (j=0 ; j<3 ; j++)
			{
				p->org[j] = org[j] + ((rand()%32)-16);
				p->vel[j] = (rand()%512)-256;
			}
		}
	}
}


void R_ParticleWat (vec3_t org, int colr)
{
	int			i, j;
	particle_t	*p;

	for (i=0 ; i<2 ; i++)
	{
		if (!free_particles)
			return;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 0.02;
		p->color = colr;
		
		{
			p->type = pt_static;
			for (j=0 ; j<3 ; j++)
			{
				p->org[j] = org[j];
				p->vel[j] = (rand()%64)-32;
			}
		}
	}
}
/*
===============
R_ParticleExplosion2

===============
*/
void R_ParticleExplosion2 (vec3_t org, int colorStart, int colorLength)
{
	int			i, j;
	particle_t	*p;
	int			colorMod = 0;

	for (i=0; i<512; i++)
	{
		if (!free_particles)
			return;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 0.3;
		p->color = colorStart + (colorMod % colorLength);
		colorMod++;

		p->type = pt_blob;
		for (j=0 ; j<3 ; j++)
		{
			p->org[j] = org[j] + ((rand()%32)-16);
			p->vel[j] = (rand()%512)-256;
		}
	}
}

/*
===============
R_BlobExplosion

===============
*/
void R_BlobExplosion (vec3_t org)
{
	int			i, j;
	particle_t	*p;

	for (i=0 ; i<1024 ; i++)
	{
		if (!free_particles)
			return;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 1 + (rand()&8)*0.05;

		if (i & 1)
		{
			p->type = pt_blob;
			p->color = 66 + rand()%6;
			for (j=0 ; j<3 ; j++)
			{
				p->org[j] = org[j] + ((rand()%32)-16);
				p->vel[j] = (rand()%512)-256;
			}
		}
		else
		{
			p->type = pt_blob2;
			p->color = 150 + rand()%6;
			for (j=0 ; j<3 ; j++)
			{
				p->org[j] = org[j] + ((rand()%32)-16);
				p->vel[j] = (rand()%512)-256;
			}
		}
	}
}



/*
===============
R_PlasmaExplosion
used by TE_TEI_PLASMAHIT
===============
*/
void R_PlasmaExplosion (vec3_t org, int colorStart, int colorLength, int count)
{
	int			i, j;
	particle_t	*p;
	int			colorMod = 0;

	for (i=0; i<count; i++)
	{
		if (!free_particles)
			return;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;
		// TODO: Implement Spark Trail type
		p->die = cl.time + 0.4;
		p->color = colorStart + (colorMod % colorLength);
		colorMod++;

		p->type = pt_decel;
		for (j=0 ; j<3 ; j++)
		{
			p->org[j] = org[j] + ((rand()%4)-2);
			p->vel[j] = (rand()%1500)-750;
		}
	}
}
/*
===============
R_RunParticleEffect



===============
*/
#ifdef PROTO
//
// Proto Particles have gravity and spew from origin making
// it a splash like effect instead of a leak like effect
// !!!!!!!!!!!!!!!!
// !HACK HACK HACK!
// !!!!!!!!!!!!!!!!
// Color overrides: Not exactly override of colors,
// but a stupid hack to allow other effects without
// doing something stupid to break the built-ins!
//
// > 256 = Smoke effect, no super velocity or gravity
// > 512 = Blood effect, splatter (todo! dynamically splatter ON THE SCREEN when touched on view then fade)
// > 768 = Rocks and debris, falling down, LONG life time
// > 1024 = !?
//
void R_RunParticleEffect (vec3_t org, vec3_t dir, int color, int count)
{
	int			i, j;
	particle_t	*p;

	for (i=0 ; i<count ; i++)
	{
		if (!free_particles)
			return;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;
//		p->splatter = 0;
		if (count == 1024)
		{	// rocket explosion
			p->die = cl.time + 5;
			p->color = ramp1[0];
			p->ramp = rand()&3;

			if (i & 1)
			{
				p->type = pt_explode;
				for (j=0 ; j<3 ; j++)
				{
					p->org[j] = org[j] + ((rand()%32)-16);
					p->vel[j] = (rand()%512)-256;
				}
			}
			else
			{
				p->type = pt_explode2;
				for (j=0 ; j<3 ; j++)
				{
					p->org[j] = org[j] + ((rand()%32)-16);
					p->vel[j] = (rand()%512)-256;
				}
			}
		}
		else
		{
			p->die = cl.time + 0.1*(rand()%8);
			//p->color = (color&~7) + (rand()&7);

			p->color = (color&~2) + (rand()&2);
					if (p->color > 64 && p->color < 79){
//					p->splatter = 3;
					p->type = pt_slowgrav;
					p->die = cl.time + 12;
					}
			if (color > 256){
				p->type = pt_slowgrav;
			p->color = p->color - 256; }
	
			else if (color > 512){
				p->type = pt_explode;
			p->color = p->color - 512; }
	
			else if (color > 768){
				p->type = pt_explode;
			p->color = p->color - 768; }
		
			else if (color > 1024){
				p->type = pt_explode;
				p->color = p->color - 1024; }
		
			else{
			p->type = pt_grav;
			
				}
			for (j=0 ; j<3 ; j++)
			{
				p->org[j] = org[j];
			//	p->vel[j] = dir[j]*15 + (rand()%300)-150;
			//		p->vel[j] = dir[j]*15 + (rand()%150)-75;
						p->vel[j] = dir[j]*((rand()%22)) + (rand()%50)-25;
			}
		}
	}
}
#else
//
// The original Quack particles
//
void R_RunParticleEffect (vec3_t org, vec3_t dir, int color, int count)
{
	int			i, j;
	particle_t	*p;

	for (i=0 ; i<count ; i++)
	{
		if (!free_particles)
			return;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;
//		p->splatter = 0;
		if (count == 1024)
		{	// rocket explosion
			p->die = cl.time + 5;
			p->color = ramp1[0];
			p->ramp = rand()&3;
			if (i & 1)
			{
				p->type = pt_explode;
				for (j=0 ; j<3 ; j++)
				{
					p->org[j] = org[j] + ((rand()%32)-16);
					p->vel[j] = (rand()%512)-256;
				}
			}
			else
			{
				p->type = pt_explode2;
				for (j=0 ; j<3 ; j++)
				{
					p->org[j] = org[j] + ((rand()%32)-16);
					p->vel[j] = (rand()%512)-256;
				}
			}
		}
		else
		{
			p->die = cl.time + 0.1*(rand()%5);
			p->color = (color&~7) + (rand()&7);
			p->type = pt_slowgrav;
			for (j=0 ; j<3 ; j++)
			{
				p->org[j] = org[j] + ((rand()&15)-8);
				p->vel[j] = dir[j]*15;// + (rand()%300)-150;
			
			}
			if (particleblood){
			if ((color > 63) && (color < 80)){
				//p->lit = 1;
				p->type = pt_blood;
				}
			}
		}
	}
}



// different alternate 'additive' method
// origin is smaller and velocity is more random
// it also fades and prefers scale over quantity (to be implemented)
// using it depends on your taste.

void R_RunWarticleEffect (vec3_t org, vec3_t dir, int color, int count)
{
	int			i, j;
	particle_t	*p;

	for (i=0 ; i<count ; i++)
	{
		if (!free_particles)
			return;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;
		if (count == 1024)
		{	// rocket explosion
			p->die = cl.time + 5;
			p->color = ramp1[0];
			p->ramp = rand()&3;
			if (i & 1)
			{
				p->type = pt_explode;
				for (j=0 ; j<3 ; j++)
				{
					p->org[j] = org[j] + ((rand()%32)-16);
					p->vel[j] = (rand()%512)-256;
				}
			}
			else
			{
				p->type = pt_explode2;
				for (j=0 ; j<3 ; j++)
				{
					p->org[j] = org[j] + ((rand()%32)-16);
					p->vel[j] = (rand()%512)-256;
				}
			}
		}
		else
		{
			p->die = cl.time + 0.2*(rand()%5);
			p->color = (color&~7) + (rand()&7);
			p->alpha = 0.5 + (rand()%6 / 6);
			p->alphavel = -1;
			p->color = color;
			p->type = pt_slowgravaddfade;
			for (j=0 ; j<3 ; j++)
			{
				//p->org[j] = org[j] + ((rand()&15)-8);
				p->org[j] = org[j];
				p->vel[j] = dir[j]*15 + (rand()%50)-25;
			}
						if (particleblood){
			if ((color > 63) && (color < 80)){
				p->type = pt_blood;
				}
			}
		}
	}
}







#endif

// --------------------
// Used by TE_TEI_SMOKE
// --------------------

void R_Smoke (vec3_t org, vec3_t dir, int color, int count)
{
	int			i, j;
	particle_t	*p;

	for (i=0 ; i<count ; i++)
	{
		if (!free_particles)
			return;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;
//		p->splatter = 0;
		if (count == 1024)
		{	// rocket explosion
			p->die = cl.time + 5;
			p->color = ramp1[0];
			p->ramp = rand()&3;
			if (i & 1)
			{
				p->type = pt_smoke;
				for (j=0 ; j<3 ; j++)
				{
					p->org[j] = org[j] + ((rand()%32)-16);
					p->vel[j] = (rand()%512)-256;
				}
			}
			else
			{
				p->type = pt_smoke;
				for (j=0 ; j<3 ; j++)
				{
					p->org[j] = org[j] + ((rand()%32)-16);
					p->vel[j] = (rand()%512)-256;
				}
			}
		}
		else
		{
			p->die = cl.time + 0.1*(rand()%5);
			p->color = (color&~7) + (rand()&7);
			p->type = pt_smoke;
			for (j=0 ; j<3 ; j++)
			{
				p->org[j] = org[j] + ((rand()%6)-3);
				p->vel[j] = (rand()%50)-25;
			}
		}
	}
}


// --------------------
// Used for the engine splash calls (by r_particle_splash)
// --------------------

void R_Splash (vec3_t org, vec3_t dir, int color, int count, int power)
{
	int			i, j;
	particle_t	*p;

	for (i=0 ; i<count ; i++)
	{
		if (!free_particles)
			return;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;
//		p->splatter = 0;
		if (count == 1024)
		{	// rocket explosion
			p->die = cl.time + 5;
			p->color = ramp1[0];
			p->ramp = rand()&3;
			if (i & 1)
			{
				p->type = pt_smoke;
				for (j=0 ; j<3 ; j++)
				{
					p->org[j] = org[j] + ((rand()%32)-16);
					p->vel[j] = (rand()%512)-256;
				}
			}
			else
			{
				p->type = pt_smoke;
				for (j=0 ; j<3 ; j++)
				{
					p->org[j] = org[j] + ((rand()%32)-16);
					p->vel[j] = (rand()%512)-256;
				}
			}
		}
		else
		{
			p->die = cl.time + 0.4*(rand()%5);
			p->color = color;
			p->type = pt_blob;
			for (j=0 ; j<3 ; j++)
			{
				p->org[j] = org[j];
				p->vel[j] = (rand()%350)-(125);
			}
		}
	}
}

/*
===============
R_LavaSplash

===============
*/
#ifdef PROTO
void R_LavaSplash (vec3_t org)
{
	// more like gib splash!
	int			i, j, k;
	particle_t	*p;
	float		vel;
	vec3_t		dir;

	for (i=-16 ; i<16 ; i++)
		for (j=-16 ; j<16 ; j++)
			for (k=0 ; k<1 ; k++)
			{
		if (!free_particles)
					return;
				p = free_particles;
				free_particles = p->next;
				p->next = active_particles;
				active_particles = p;

				p->die = cl.time + 2 + (rand()&31) * 4.02;
				p->color = 68 + (rand()&7);
					p->type = pt_grav;

				dir[0] = j*448 + (rand()&27);
				dir[1] = i*448 + (rand()&27);
				dir[2] = 2356;

				p->org[0] = org[0];
				p->org[1] = org[1];
				p->org[2] = org[2];
//				p->splatter = 3; // stick to the walls
				VectorNormalize (dir);
				vel = 50 + (rand()&63);
				VectorScale (dir, vel, p->vel);

				}
		
}


#else

void R_LavaSplash (vec3_t org)
{
	int			i, j, k;
	particle_t	*p;
	float		vel;
	vec3_t		dir;

	for (i=-16 ; i<16 ; i++)
		for (j=-16 ; j<16 ; j++)
			for (k=0 ; k<1 ; k++)
			{
				if (!free_particles)
					return;
				p = free_particles;
				free_particles = p->next;
				p->next = active_particles;
				active_particles = p;

				p->die = cl.time + 2 + (rand()&31) * 0.02;
				p->color = 224 + (rand()&7);
				p->type = pt_slowgrav;

				dir[0] = j*8 + (rand()&7);
				dir[1] = i*8 + (rand()&7);
				dir[2] = 256;

				p->org[0] = org[0] + dir[0];
				p->org[1] = org[1] + dir[1];
				p->org[2] = org[2] + (rand()&63);

				VectorNormalize (dir);
				vel = 50 + (rand()&63);
				VectorScale (dir, vel, p->vel);
			}
}
#endif


/*
===============
R_Blood
===============
*/
void R_Blood(int count, vec3_t mins, vec3_t maxs, vec3_t vel_mins, vec3_t vel_maxs)
{
	const int color = 73;
	int idist[3], vdist[3];
	int i, j;
	particle_t *p;

	for (i = 0; i < 3; i++)
	{
		idist[i] = (int)(maxs[i] - mins[i] + 0.1f);
		vdist[i] = (int)(vel_maxs[i] - vel_mins[i] + 0.1f);

		if (idist[i] < 1)
			idist[i] = 1;
		if (vdist[i] < 1)
			vdist[i] = 1;
	}

	for (i = 0; i < count; i++)
	{
		if (!free_particles)
			break;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		//p->die = cl.time + 0.1*(rand()%5);
		p->die = cl.time + 0.1*(rand()%65);
		p->color = (float)((color&~7) + (rand()&7));
		//p->splatter = 3;
		//p->lit = 1;
		p->type = pt_blood;
		for (j = 0; j < 3; j++)
		{
			p->org[j] = mins[j] + rand() % idist[j] + ((rand()&15)-8);
			p->vel[j] = vel_mins[j] + rand() % vdist[j];
		}
	}
}




/*
===============
R_TeleportSplash

===============
*/
void R_TeleportSplash (vec3_t org)
{
	int			i, j, k;
	particle_t	*p;
	float		vel;
	vec3_t		dir;

	for (i=-16 ; i<16 ; i+=4)
		for (j=-16 ; j<16 ; j+=4)
			for (k=-24 ; k<32 ; k+=4)
			{
				if (!free_particles)
					return;
				p = free_particles;
				free_particles = p->next;
				p->next = active_particles;
				active_particles = p;

				p->die = cl.time + 0.2 + (rand()&7) * 0.02;
				p->color = 7 + (rand()&7);
				p->type = pt_slowgrav;

				dir[0] = j*8;
				dir[1] = i*8;
				dir[2] = k*8;

				p->org[0] = org[0] + i + (rand()&3);
				p->org[1] = org[1] + j + (rand()&3);
				p->org[2] = org[2] + k + (rand()&3);

				VectorNormalize (dir);
				vel = 50 + (rand()&63);
				VectorScale (dir, vel, p->vel);
			}
}

void R_RocketTrail (vec3_t start, vec3_t end, int type)
{
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;
	int			dec;
	static int	tracercount;

	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);
	if (type < 128)
		dec = 3;
	else 	if (type == 7)
		dec = 1;
	else
	{
		dec = 1;
		type -= 128;
	}

	while (len > 0)
	{
		len -= dec;

		if (!free_particles)
			return;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		VectorCopy (vec3_origin, p->vel);
		p->die = cl.time + 2;

		switch (type)
		{
			case 0:	// rocket trail
				p->ramp = (rand()&3);
				p->color = ramp3[(int)p->ramp];
				p->type = pt_fire;
				for (j=0 ; j<3 ; j++)
					p->org[j] = start[j] + ((rand()%6)-3);
				break;

			case 1:	// smoke smoke
				p->ramp = (rand()&3) + 2;
				p->color = ramp3[(int)p->ramp];
				p->type = pt_fire;
				for (j=0 ; j<3 ; j++)
					p->org[j] = start[j] + ((rand()%6)-3);
				break;

			case 2:	// blood
				p->type = pt_grav;
				p->color = 67 + (rand()&3);
				for (j=0 ; j<3 ; j++)
					p->org[j] = start[j] + ((rand()%6)-3);
				if (particleblood){
					//p->lit = 1;
					p->type = pt_blood;
				}
				break;

			case 3:
			case 5:	// tracer
				p->die = cl.time + 0.5;
				p->type = pt_static;
				if (type == 3)
					p->color = 52 + ((tracercount&4)<<1);
				else
					p->color = 230 + ((tracercount&4)<<1);

				tracercount++;

				VectorCopy (start, p->org);
				if (tracercount & 1)
				{
					p->vel[0] = 30*vec[1];
					p->vel[1] = 30*-vec[0];
				}
				else
				{
					p->vel[0] = 30*-vec[1];
					p->vel[1] = 30*vec[0];
				}
				break;

			case 4:	// slight blood
				p->type = pt_grav;
				p->color = 67 + (rand()&3);
				for (j=0 ; j<3 ; j++)
					p->org[j] = start[j] + ((rand()%6)-3);
				if (particleblood){
					//p->lit = 1;
					p->type = pt_blood;
				}
				len -= 3;
				break;

			case 6:	// voor trail
				p->color = 9*16 + 8 + (rand()&3);
				p->type = pt_static;
				p->die = cl.time + 0.3;
				for (j=0 ; j<3 ; j++)
					p->org[j] = start[j] + ((rand()&15)-8);
				break;
			case 7:	// rail trail smoke
				p->ramp = (rand()&3) + 2;
				p->color = ramp3[(int)p->ramp];
				p->alpha = 1;
				p->alphavel = -3.7 - rand()%1;
				p->type = pt_staticfade;
				for (j=0 ; j<3 ; j++)
					p->org[j] = start[j] + ((rand()%6)-3);
				break;
		}
	

		VectorAdd (start, vec, start);
	}
}

// LEI - lame particle beam
// (TODO: Quake2 polygon beam)
void R_ParticleBeam (vec3_t start, vec3_t end, int thecol)
{
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;
	int			dec;
	static int	tracercount;
	int	wat;
	//int	thecol;

	

	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);
	dec = 1;

	while (len > 0)
	{
		len -= dec;

		if (!free_particles)
			return;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;
		
		VectorCopy (vec3_origin, p->vel);
		p->die = cl.time;

		
		wat = rand()&2;
		{
			if (wat == 2){	// Outer Glow
				p->color = thecol;
				p->type = pt_staticfadeadd;
				p->die = cl.time + 2;
				p->alpha = 0.6;
				p->alphavel = -0.7;
				for (j=0 ; j<3 ; j++)
					p->org[j] = start[j] + ((rand()%4)-2);
			}
			else 
			{	// Inner Line
				p->color = thecol;
				p->type = pt_staticfadeadd;
				p->die = cl.time + 2;
				p->alpha = 1;
				p->alphavel = -0.9;
				for (j=0 ; j<3 ; j++)
					p->org[j] = start[j];
			}
		}


		VectorAdd (start, vec, start);
	}
}


/*
===============
R_RailTrail
Used by TE_RAILTRAIL
i'm a quake2 function short and stout
===============
*/


void MakeNormalVectors (vec3_t forward, vec3_t right, vec3_t up)
{
	float		d;

	// this rotate and negat guarantees a vector
	// not colinear with the original
	right[1] = -forward[0];
	right[2] = forward[1];
	right[0] = forward[2];

	d = DotProduct (right, forward);
	VectorMA (right, -d, forward, right);
	VectorNormalize (right);
	CrossProduct (right, forward, up);
}



void R_RailTrail (vec3_t start, vec3_t end)
{
vec3_t		move;
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;
	float		dec;
	vec3_t		right, up;
	int			i;
	float		d, c, s;
	vec3_t		dir;
	byte		clr = 40;

	VectorCopy (start, move);
	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);

	MakeNormalVectors (vec, right, up);

	for (i=0 ; i<len ; i++)
	{
		if (!free_particles)
			return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;
		
		p->die = cl.time + 2.0;
		VectorClear (p->vel);

		d = i * 0.1;
		c = cos(d);
		s = sin(d);

		VectorScale (right, c, dir);
		VectorMA (dir, s, up, dir);
		p->type = pt_staticfadeadd;
		p->alpha = 1.0;
	//	p->alphavel = -1.0 / (1+frand()*0.2);
		p->alphavel = -1.0;

		p->color = clr + (rand()&7);
		for (j=0 ; j<3 ; j++)
		{
			p->org[j] = move[j] + dir[j]*3;
			p->vel[j] = dir[j]*6;
		}

		VectorAdd (move, vec, move);
	}

	dec = 0.75;
	VectorScale (vec, dec, vec);
	VectorCopy (start, move);

/*	while (len > 0)
	{
		len -= dec;

		if (!free_particles)
			return;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;
//		VectorScale (right, c, dir);
//		VectorMA (dir, s, up, dir);
		p->die = cl.time + 2.1;
		VectorClear (p->vel);
		p->type = pt_fire;
		p->alpha = 1.0;
	//	p->alphavel = -1.0 / (0.6+frand()*0.2);
	//	p->color = 0 + rand()&15;
			p->color = 7;

		for (j=0 ; j<3 ; j++)
		{
			p->org[j] = move[j] + rand()*3;
			p->vel[j] = rand()*3;
		//	p->vel[j] = 0;
		}

		VectorAdd (move, vec, move);
	}*/

}


/*
===============
R_DrawParticles
===============
*/
extern cvar_t	*sv_gravity;
#ifdef PROTO
void R_DrawParticles (void)
{
	particle_t		*p, *kill;
	float			grav, grav2, percent;
	int				i;
	float			time2, time3;
	float			time1;
	float			dvel;
	float			frametime;

// hexen 2
		vec3_t		diff;
		qboolean	in_solid;

#ifdef GLQUAKE
	vec3_t			up, right;
	float			scale;
/*
    GL_Bind(particletexture);
	glEnable (GL_BLEND);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glBegin (GL_TRIANGLES);

	VectorScale (vup, 1.5, up);
	VectorScale (vright, 1.5, right);
	// additive hack
*/
	GL_Bind(particletexture);
	glEnable (GL_BLEND);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glBegin (GL_TRIANGLES);

	VectorScale (vup, 8.5, up);
	VectorScale (vright, 8.5, right);

#else
	D_StartParticles ();

	VectorScale (vright, xscaleshrink, r_pright);
	VectorScale (vup, yscaleshrink, r_pup);
	VectorCopy (vpn, r_ppn);
#endif
	frametime = cl.time - cl.oldtime;
	time3 = frametime * 15;
	time2 = frametime * 10; // 15;
	time1 = frametime * 5;
	grav = frametime * sv_gravity->value * 0.05;
	dvel = 4*frametime;

	for ( ;; )
	{
		kill = active_particles;
		if (kill && kill->die < cl.time)
		{
			active_particles = kill->next;
			kill->next = free_particles;
			free_particles = kill;
			continue;
		}
		break;
	}

	for (p=active_particles ; p ; p=p->next)
	{
		for ( ;; )
		{
			kill = p->next;
			if (kill && kill->die < cl.time)
			{
				p->next = kill->next;
				kill->next = free_particles;
				free_particles = kill;
				continue;
			}
			break;
		}

#ifdef GLQUAKE
		// hack a scale up to keep particles from disapearing
		scale = (p->org[0] - r_origin[0])*vpn[0] + (p->org[1] - r_origin[1])*vpn[1]
			+ (p->org[2] - r_origin[2])*vpn[2];
		if (scale < 20)
			scale = 1;
		else
			scale = 1 + scale * 0.004;
		glColor3ubv ((byte *)&d_8to24table[(int)p->color]);
		glTexCoord2f (0,0);
		glVertex3fv (p->org);
		glTexCoord2f (1,0);
		glVertex3f (p->org[0] + up[0]*scale, p->org[1] + up[1]*scale, p->org[2] + up[2]*scale);
		glTexCoord2f (0,1);
		glVertex3f (p->org[0] + right[0]*scale, p->org[1] + right[1]*scale, p->org[2] + right[2]*scale);
#else
		D_DrawParticle (p);
#endif

// PARTICLE COLLISION CODE FROM HEXEN II! (It used to be for snow now it's for everything!)
// BEIGN
		if (p->splatter >= 1)
			{
			// FIXME: If flake going fast enough, can go through,
			//	  do a check in increments ot 10, max?
			// if not in_bounds Get length of diff, add in
			// increments of 4 & check solid
				mleaf_t		*l;


				{
					VectorScale(p->vel, frametime, diff);
					VectorAdd(p->org, diff, p->org);
				}
			

				{

					// WHERE THE HITTING HAPPENS
					// if hit solid, go to last position,
					// no velocity, fade out.
					l = Mod_PointInLeaf (p->org, cl.worldmodel);
					if (l->contents != CONTENTS_EMPTY) // || in_solid == true
					{
					
						{
						// still have small prob of snow melting on emitter
							VectorScale(diff, 0.2, p->vel);
							i = 6;
							while (l->contents != CONTENTS_EMPTY)
							{
								p->org[0] -= p->vel[0];
								p->org[1] -= p->vel[1];
								p->org[2] -= p->vel[2];
								i--; //no infinite loops
								if (!i)
								{
									p->die = -1;	//should never happen now!
									break;
								}
								l = Mod_PointInLeaf (p->org, cl.worldmodel);
							}
							p->vel[0] = p->vel[1] = p->vel[2] = 0;
							p->ramp = 0;
//							if (p->splatter = 3)
//									p->type = pt_static; // apply blood fall gravity.
						}
					}
				}
				

			}
	

			// END CODE of stupid hexen 2 particle collision
		//else
		

	// make 'em move
		else
		{
			p->org[0] += p->vel[0] * frametime;
			p->org[1] += p->vel[1] * frametime;
			p->org[2] += p->vel[2] * frametime;
		}
	
		


		switch (p->type)
		{
//			p->splatter = 0;
		case pt_static:
			break;
		case pt_fire:
			p->ramp += time1;
			if (p->ramp >= 6)
				p->die = -1;
			else
				p->color = ramp3[(int)p->ramp];
			p->vel[2] += grav;
			break;

		case pt_explode:
			p->ramp += time2;
//			p->splatter = 0;
			if (p->ramp >=8)
				p->die = -1;
			else
				p->color = ramp1[(int)p->ramp];
			for (i=0 ; i<3 ; i++)
				p->vel[i] += p->vel[i]*dvel;
			p->vel[2] -= grav;
			break;

		case pt_explode2:
			p->ramp += time3;
//			p->splatter = 0;
			if (p->ramp >=8)
				p->die = -1;
			else
				p->color = ramp2[(int)p->ramp];
			for (i=0 ; i<3 ; i++)
				p->vel[i] -= p->vel[i]*frametime;
			p->vel[2] -= grav;
			break;

		case pt_blob:
			for (i=0 ; i<3 ; i++)
				p->vel[i] += p->vel[i]*dvel;
			p->vel[2] -= grav;
			break;

		case pt_blob2:
			for (i=0 ; i<2 ; i++)
				p->vel[i] -= p->vel[i]*dvel;
			p->vel[2] -= grav;
			break;

		case pt_grav:
			p->vel[2] -= grav * 80;
			break;

		case pt_slowgrav:
			p->vel[2] -= (grav / 3);
			break;

		case pt_fastgrav:
			p->vel[2] -= grav * 1;
			break;

		case pt_staticfade:
			p->alpha + frametime*p->alphavel;

			if (p->alpha <= 0)
				p->die = -1;
			break;
	
		}
		
	}
#ifdef GLQUAKE
	glEnd ();
	glDisable (GL_BLEND);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
#else
	D_EndParticles ();
#endif

}
#else


void R_DrawParticles (void)
{
	particle_t		*p, *kill;
	float			grav;
	int				i;
	float			time2, time3;
	float			time1;
	float			dvel;
	float			frametime;
// hexen 2
			mleaf_t		*l;
		vec3_t		diff;
		vec3_t		below;
		qboolean	in_solid;

#ifdef GLQUAKE
	vec3_t			up, right;
	float			scale;
	GL_Bind(particletexture);
	glEnable (GL_BLEND);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glBegin (GL_TRIANGLES);

	VectorScale (vup, 8.5, up);
	VectorScale (vright, 8.5, right);

#else
	D_StartParticles ();

	VectorScale (vright, xscaleshrink, r_pright);
	VectorScale (vup, yscaleshrink, r_pup);
	VectorCopy (vpn, r_ppn);
#endif
	frametime = cl.time - cl.oldtime;
	time3 = frametime * 15;
	time2 = frametime * 10; // 15;
	time1 = frametime * 5;
	grav = frametime * sv_gravity->value * 0.05;
	dvel = 4*frametime;

	for ( ;; )
	{
		kill = active_particles;
		if (kill && kill->die < cl.time)
		{
						
			active_particles = kill->next;
//			kill->splatter = 0;
			kill->next = free_particles;
			
			free_particles = kill;
			continue;
		}
		break;
	}

	for (p=active_particles ; p ; p=p->next)
	{
		for ( ;; )
		{
			kill = p->next;
			if (kill && kill->die < cl.time)
			{
//				p->splatter = 0;
				p->next = kill->next;
				kill->next = free_particles;
				free_particles = kill;
				continue;
			}
			break;
		}

#ifdef GLQUAKE
		// hack a scale up to keep particles from disapearing
		scale = (p->org[0] - r_origin[0])*vpn[0] + (p->org[1] - r_origin[1])*vpn[1]
			+ (p->org[2] - r_origin[2])*vpn[2];
		if (scale < 20)
			scale = 1;
		else
			scale = 1 + scale * 0.004;
		glColor3ubv ((byte *)&d_8to24table[(int)p->color]);
		glTexCoord2f (0,0);
		glVertex3fv (p->org);
		glTexCoord2f (1,0);
		glVertex3f (p->org[0] + up[0]*scale, p->org[1] + up[1]*scale, p->org[2] + up[2]*scale);
		glTexCoord2f (0,1);
		glVertex3f (p->org[0] + right[0]*scale, p->org[1] + right[1]*scale, p->org[2] + right[2]*scale);
#else



	if (rmap_ready)
		p->color= (int)(coltranslate[(int)p->color]);



		// leilei - particle alpha blending
//		if(p->alpha < 0.1)
		//		p->alpha = 1;
		// very, very, very stupid way to do this but keeps asm (?) happy


	if (p->blend > 0)
		D_DrawParticle_Add (p);	
		else {
			if (p->alpha < 0.33f)
		D_DrawParticle_A33 (p);
		else if (p->alpha < 0.66f)
		D_DrawParticle_A66 (p);

//		else if (p->lit)
//		D_DrawParticleLit (p);
		else
		D_DrawParticle (p);	
		}
#endif// PARTICLE COLLISION CODE FROM HEXEN II! (It used to be for snow now it's for everything!)
// BEIGN

		//else
		

	// make 'em move
			p->org[0] += p->vel[0] * frametime;
			p->org[1] += p->vel[1] * frametime;
			p->org[2] += p->vel[2] * frametime;
		

		switch (p->type)
		{
		case pt_static:
			p->alpha = 1;p->blend = 0;p->lit = 0;
			break;
		case pt_fire:
			p->alpha = 1;p->blend = 0;p->lit = 0;
			p->ramp += time1;
			if (p->ramp >= 6)
				p->die = -1;
			else
				p->color = ramp3[(int)p->ramp];
			p->vel[2] += grav;
			break;

		case pt_explode:
			p->alpha = 1;p->blend = 0;p->lit = 0;
			p->ramp += time2;
//			p->splatter = 0;
			if (p->ramp >=8)
				p->die = -1;
			else
				p->color = ramp1[(int)p->ramp];
			for (i=0 ; i<3 ; i++)
				p->vel[i] += p->vel[i]*dvel;
			p->vel[2] -= grav;
			break;

		case pt_explode2:
			p->alpha = 1;p->blend = 0;p->lit = 0;
			p->ramp += time3;
//			p->splatter = 0;
			if (p->ramp >=8)
				p->die = -1;
			else
				p->color = ramp2[(int)p->ramp];
			for (i=0 ; i<3 ; i++)
				p->vel[i] -= p->vel[i]*frametime;
			p->vel[2] -= grav;
			break;

		case pt_blob:
			p->alpha = 1;p->blend = 0;p->lit = 0;
			for (i=0 ; i<3 ; i++)
				p->vel[i] += p->vel[i]*dvel;
			p->vel[2] -= grav;
			break;

		case pt_blob2:
			p->alpha = 1;p->blend = 0;p->lit = 0;
			for (i=0 ; i<2 ; i++)
				p->vel[i] -= p->vel[i]*dvel;
			p->vel[2] -= grav;
			break;

		case pt_grav:
			p->alpha = 1;p->blend = 0;	p->lit = 0;
			p->vel[2] -= grav * 12;
			break;
// huge blood crap
		case pt_blood:
				l = Mod_PointInLeaf (p->org, cl.worldmodel);
						if (l->contents != CONTENTS_EMPTY) // || in_solid == true
					{
								p->type = pt_staticfade;
							p->die = p->die + 6;
							p->alphavel = -0.1 * (rand()&8);
							p->vel[0] = 0;
							p->vel[1] = 0;
							p->vel[2] = 0;
					}
								else
								{
		//	p->blend = 0;
			p->lit = 1;
		//		p->alpha = 0.5;
			p->vel[2] -= grav * 6;
		
								}
				
	/*

				{
					VectorScale(p->vel, frametime, diff);
					VectorAdd(p->org, diff, p->org);
				}
			

				{

					// WHERE THE HITTING HAPPENS
					// if hit solid, go to last position,
					// no velocity, fade out.
					l = Mod_PointInLeaf (p->org, cl.worldmodel);
					if (l->contents != CONTENTS_EMPTY) // || in_solid == true
					{
					
					{
						// still have small prob of snow melting on emitter
							VectorScale(diff, 0.2, p->vel);
							i = 6;
							while (l->contents != CONTENTS_EMPTY)
							{
								p->org[0] -= p->vel[0];
								p->org[1] -= p->vel[1];
								p->org[2] -= p->vel[2];
										
								i--; //no infinite loops
								if (!i)
								{
									p->die = -1;	//should never happen now!
									break;
								}
								l = Mod_PointInLeaf (p->org, cl.worldmodel);
							}
							//	p->vel[2] -= grav;
							//	if (rand()&8 < 3)
							{
			
							
							p->alpha = frametime - p->die;
							p->blend = 0;
							p->vel[2] -= grav;
						}
						//		else
						//			p->vel[0] = p->vel[1] = p->vel[2] = 0;
							p->ramp = 0;
						//	if (p->splatter == 6)
							{
						//			p->type = pt_static; // apply blood fall gravity.
					//	 if (rand() & 5)	p->type = pt_slowgrav; // apply blood fall gravity.
						//	p->die = p->die + 5; // live a little more
												// Just Drop Down. drip!
					below[0] = p->org[0] - 16;
					below[1] = p->org[1];
					below[2] = p->org[2];
					l = Mod_PointInLeaf (below, cl.worldmodel);
					if (l->contents == CONTENTS_EMPTY) 
					{
						if (rand()&8 < 3){
							p->die = p->die + 6;
							p->alpha = 1;

							
							p->blend = 0;
								p->vel[2] -= grav * 44;
						}
					
					}
						

					}
				
						}
					}
			
				}
				
	*/	
			
	

			// END CODE of stupid hexen 2 particle collision
			break;

		case pt_slowgrav:
			p->alpha = 1;p->blend = 0;p->lit = 0;
			p->vel[2] -= grav;
			break;
		case pt_fastgrav:
			p->alpha = 1;p->blend = 0;p->lit = 0;
			p->vel[2] -= grav * 15;
			break;
		case pt_smoke:
			p->alpha = 1;p->blend = 2; p->lit = 0;	// this is additive in dp!
			p->vel[2] += grav;
			break;
		case pt_add:
			if(p->alpha < 0.02)
					p->alpha = 1;
			else
			p->alpha -= 0.5;
			p->alpha = 0.5;
			p->blend = 2;p->lit = 0;
		break;
		case pt_decel:
			p->alpha = 1;
			p->blend = 0;p->lit = 0;
			for (i=0 ; i<3 ; i++)
				p->vel[i] += p->vel[i]* -dvel;
			break;
		case pt_staticfade:
			p->alpha += frametime*p->alphavel;
			//	p->alpha -= cl.time*(0.001 +(rand() % 6 * 0.006));
			p->blend = 0;p->lit = 0;
			if (p->alpha <= 0)
				p->die = -1;
			break;
		case pt_staticfadeadd:
			p->alpha += frametime*p->alphavel;
			p->blend = 2;p->lit = 0;
			if (p->alpha <= 0)
				p->die = -1;
			break;
		case pt_slowgravaddfade:
			p->alpha += frametime*p->alphavel;
			p->blend = 2;p->lit = 0;
			p->vel[2] -= grav;
			if (p->alpha <= 0)
				p->die = -1;
			break;
		}
	}

#ifdef GLQUAKE
	glEnd ();
	glDisable (GL_BLEND);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
#else
	D_EndParticles ();
#endif

}


#endif