/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_surf.c: surface-related refresh code

#include "quakedef.h"
#include "r_local.h"

drawsurf_t	r_drawsurf;

int				lightleft, sourcesstep, blocksize, sourcetstep;
int				lightdelta, lightdeltastep;
int				lightright, lightleftstep, lightrightstep, blockdivshift;


int				lightlefta[3], sourcesstep, blocksize, sourcetstep;
int				lightdelta, lightdeltastep;
int				lightrighta[3], lightleftstepa[3], lightrightstepa[3], blockdivshift;



// vavoom
int				lightrleft, lightrright, lightrleftstep, lightrrightstep;
int				lightgleft, lightgright, lightgleftstep, lightgrightstep;
int				lightbleft, lightbright, lightbleftstep, lightbrightstep;

unsigned		blockdivmask;
void			*prowdestbase;
unsigned char	*pbasesource;
int				surfrowbytes;	// used by ASM files



int				r_stepback;
int				r_lightwidth;
int				r_numhblocks, r_numvblocks;
unsigned char	*r_source, *r_sourcemax;
int		*r_lightptr;
int		blocklights[18*18*3]; // LordHavoc: .lit support (*3 for RGB)


void R_DrawSurfaceBlock8_mip0 (void);
void R_DrawSurfaceBlock8_mip1 (void);
void R_DrawSurfaceBlock8_mip2 (void);
void R_DrawSurfaceBlock8_mip3 (void);

void R_DrawSurfaceBlock8Fast_mip0 (void);
void R_DrawSurfaceBlock8Fast_mip1 (void);
void R_DrawSurfaceBlock8Fast_mip2 (void);
void R_DrawSurfaceBlock8Fast_mip3 (void);
// vavoom



#ifdef id386rgb
void D_DrawSurfaceBlock8RGBASM_mip0(void);
void D_DrawSurfaceBlock8RGBASM_mip1(void);
void D_DrawSurfaceBlock8RGBASM_mip2(void);
void D_DrawSurfaceBlock8RGBASM_mip3(void);
#endif

void D_DrawSurfaceBlock8RGBX_mip0(void);
void D_DrawSurfaceBlock8RGBX_mip1(void);
void D_DrawSurfaceBlock8RGBX_mip2(void);
void D_DrawSurfaceBlock8RGBX_mip3(void);

void D_DrawSurfaceBlock8RGBXD_mip0(void);
void D_DrawSurfaceBlock8RGBXD_mip1(void);
void D_DrawSurfaceBlock8RGBXD_mip2(void);
void D_DrawSurfaceBlock8RGBXD_mip3(void);


static void	(*surfmiptable[4])(void) = {
	R_DrawSurfaceBlock8_mip0,
	R_DrawSurfaceBlock8_mip1,
	R_DrawSurfaceBlock8_mip2,
	R_DrawSurfaceBlock8_mip3
};


static void	(*surfmiptable8RGBX[4])(void) =
{
#if id386rgb
	R_DrawSurfaceBlock8RGBASM_mip0,
	R_DrawSurfaceBlock8RGBASM_mip1,
	R_DrawSurfaceBlock8RGBASM_mip2,
	R_DrawSurfaceBlock8RGBASM_mip3
#else
	R_DrawSurfaceBlock8RGBX_mip0,
	R_DrawSurfaceBlock8RGBX_mip1,
	R_DrawSurfaceBlock8RGBX_mip2,
	R_DrawSurfaceBlock8RGBX_mip3
#endif
};

static void	(*surfmiptable8RGBXD[4])(void) =
{

	R_DrawSurfaceBlock8RGBXD_mip0,
	R_DrawSurfaceBlock8RGBXD_mip1,
	R_DrawSurfaceBlock8RGBXD_mip2,
	R_DrawSurfaceBlock8RGBXD_mip3
};


// leilei - fast versions of these
// are just the same, sans lightstepping
// which actually brings a cool speed boost
#if !id386
static void	(*surfmiptablefast[4])(void) = {
	R_DrawSurfaceBlock8_mip0,
	R_DrawSurfaceBlock8_mip1,
	R_DrawSurfaceBlock8_mip2,
	R_DrawSurfaceBlock8_mip3
};

// our c versions suck for the time being.
#else

static void	(*surfmiptablefast[4])(void) = {
	R_DrawSurfaceBlock8Fast_mip0,
	R_DrawSurfaceBlock8Fast_mip1,
	R_DrawSurfaceBlock8Fast_mip2,
	R_DrawSurfaceBlock8Fast_mip3
};

#endif



/*
===============
R_AddDynamicLights
===============
*/
void R_AddDynamicLights (void)
{
	msurface_t *surf;
	int			lnum;
	int			sd, td;
	float		dist, rad, minlight;
	vec3_t		impact, local;
	vec3_t      origin_for_ent; // mankrip - dynamic lights on moving brush models fix
	int			s, t;
	int			i;
	int			smax, tmax;
	mtexinfo_t	*tex;
	
	
	surf = r_drawsurf.surf;
	smax = (surf->extents[0]>>4)+1;
	tmax = (surf->extents[1]>>4)+1;
	tex = surf->texinfo;

	for (lnum=0 ; lnum<MAX_DLIGHTS ; lnum++)
	{
		if ( !(surf->dlightbits & (1<<lnum) ) )
			continue;		// not lit by this light

		rad = cl_dlights[lnum].radius;
	    
		VectorSubtract (cl_dlights[lnum].origin, currententity->origin, origin_for_ent); // mankrip - dynamic lights on moving brush models fix
        dist = DotProduct (origin_for_ent, surf->plane->normal); // mankrip - dynamic lights on moving brush models fix - edited

		rad -= fabs(dist);

		minlight = cl_dlights[lnum].minlight;

		if (rad < minlight)
			continue;

		minlight = rad - minlight;

		for (i=0 ; i<3 ; i++)
		{
			impact[i] = origin_for_ent[i] - // mankrip - dynamic lights on moving brush models fix - edited
					surf->plane->normal[i]*dist;

		}

		local[0] = DotProduct (impact, tex->vecs[0]) + tex->vecs[0][3];
		local[1] = DotProduct (impact, tex->vecs[1]) + tex->vecs[1][3];

		local[0] -= surf->texturemins[0];
		local[1] -= surf->texturemins[1];

		for (t = 0 ; t<tmax ; t++)
		{
			td = local[1] - t*16;
			if (td < 0)
				td = -td;
			for (s=0 ; s<smax ; s++)
			{
				sd = local[0] - s*16;
				if (sd < 0)
					sd = -sd;
				if (sd > td)
					dist = sd + (td>>1);
				else
					dist = td + (sd>>1);
				if (dist < minlight)

				{
					unsigned temp;
					temp = (rad - dist)*256;
					// Lei - Fire Light Effect 	
					// TODO: Make this somehow usable
				//temp = (rad - (dist - rand()%92))*256;

				// Star Hack
				//	temp = (rad)*256;

					i = t*smax + s;
					if (!cl_dlights[lnum].dark)
						blocklights[i] += temp;
					else
					{
						if (blocklights[i] > temp)
							blocklights[i] -= temp;
						else
							blocklights[i] = 0;
					}
				}

	


			}
		}
	}
}

void R_AddDynamicLightsRGB ()
{
	msurface_t *surf;

	int		lnum;
	int		sd, td;
	float		dist, rad, minlight;
	vec3_t		impact, local;
	int		s, t;
	int		i;
	int		smax, tmax;
	mtexinfo_t	*tex;
	vec3_t      origin_for_ent; // mankrip - dynamic lights on moving brush models fix
	
// 2001-09-11 Colored lightning by LordHavoc/Sarcazm/Maddes  start
	float		cred, cgreen, cblue, brightness;
	unsigned	*bl;
// 2001-09-11 Colored lightning by LordHavoc/Sarcazm/Maddes  end
	surf = r_drawsurf.surf;
	smax = (surf->extents[0]>>4)+1;
	tmax = (surf->extents[1]>>4)+1;
	tex = surf->texinfo;

	for (lnum=0 ; lnum<MAX_DLIGHTS ; lnum++)
	{
		if ( !(surf->dlightbits & (1<<lnum) ) )
			continue;		// not lit by this light

		rad = cl_dlights[lnum].radius;

		VectorSubtract (cl_dlights[lnum].origin, currententity->origin, origin_for_ent); // mankrip - dynamic lights on moving brush models fix
        dist = DotProduct (origin_for_ent, surf->plane->normal); // mankrip - dynamic lights on moving brush models fix - edited


		rad -= fabs(dist);
		minlight = cl_dlights[lnum].minlight;
		if (rad < minlight)
			continue;
		minlight = rad - minlight;

		for (i=0 ; i<3 ; i++)
		{
			impact[i] = origin_for_ent[i] - // mankrip - dynamic lights on moving brush models fix - edited
					surf->plane->normal[i]*dist;
		}

		local[0] = DotProduct (impact, tex->vecs[0]) + tex->vecs[0][3];
		local[1] = DotProduct (impact, tex->vecs[1]) + tex->vecs[1][3];

		local[0] -= surf->texturemins[0];
		local[1] -= surf->texturemins[1];

// 2001-09-11 Colored lightning by LordHavoc/Sarcazm/Maddes  start
		cred = cl_dlights[lnum].color[0] * 256.0f;
		cgreen = cl_dlights[lnum].color[1] * 256.0f;
		cblue = cl_dlights[lnum].color[2] * 256.0f;

		bl = blocklights;
// 2001-09-11 Colored lightning by LordHavoc/Sarcazm/Maddes  end

		for (t = 0 ; t<tmax ; t++)
		{
			td = local[1] - t*16;
			if (td < 0)
				td = -td;
			for (s=0 ; s<smax ; s++)
			{
				sd = local[0] - s*16;
				if (sd < 0)
					sd = -sd;
				if (sd > td)
					dist = sd + (td>>1);
				else
					dist = td + (sd>>1);
				if (dist < minlight)
// 2001-09-11 Colored lightning by LordHavoc/Sarcazm/Maddes  start
//					blocklights[t*smax + s] += (rad - dist)*256;
				{
					brightness = rad - dist;
					bl[0] += (int) (brightness * cred);
					bl[1] += (int) (brightness * cgreen);
					bl[2] += (int) (brightness * cblue);
				}
				bl += 3;
// 2001-09-11 Colored lightning by LordHavoc/Sarcazm/Maddes  end
			}
		}
	}
}

// leilei - shadowhack continued

/*
===============
R_AddDynamicLights
===============
*/
void R_AddShadows (void)
{
	msurface_t *surf;
	int			lnum;
	int			sd, td;
	float		dist, rad, minlight;
	vec3_t		impact, local;
	int			s, t;
	int			i;
	int			smax, tmax;
	mtexinfo_t	*tex;
	vec3_t      origin_for_ent; // mankrip - dynamic lights on moving brush models fix
	
	surf = r_drawsurf.surf;
	smax = (surf->extents[0]>>4)+1;
	tmax = (surf->extents[1]>>4)+1;
	tex = surf->texinfo;

	for (lnum=0 ; lnum<MAX_SHADOWS ; lnum++)
	{
		if ( !(surf->shadowbits & (1<<lnum) ) )
			continue;		// not lit by this light

		rad = cl_shadows[lnum].radius;
		 VectorSubtract (cl_shadows[lnum].origin, currententity->origin, origin_for_ent); // mankrip - dynamic lights on moving brush models fix
		 	
		 dist = DotProduct (origin_for_ent, surf->plane->normal) - // mankrip - dynamic lights on moving brush models fix - edited
				surf->plane->dist;

		rad -= fabs(dist);
		minlight = cl_shadows[lnum].minlight;
		if (rad < minlight)
			continue;
		minlight = rad - minlight;

		for (i=0 ; i<3 ; i++)
		{
		 impact[i] = origin_for_ent[i] - // mankrip - dynamic lights on moving brush models fix - edited
			 			surf->plane->normal[i]*dist;
		}

		local[0] = DotProduct (impact, tex->vecs[0]) + tex->vecs[0][3];
		local[1] = DotProduct (impact, tex->vecs[1]) + tex->vecs[1][3];

		local[0] -= surf->texturemins[0];
		local[1] -= surf->texturemins[1];

		for (t = 0 ; t<tmax ; t++)
		{
			td = local[1] - t*16;
			if (td < 0)
				td = -td;
			for (s=0 ; s<smax ; s++)
			{
				sd = local[0] - s*16;
				if (sd < 0)
					sd = -sd;
				if (sd > td)
					dist = sd + (td>>1);
				else
					dist = td + (sd>>1);
				if (dist < minlight)

				{
					unsigned temp;
					temp = (rad - dist)*256;
					i = t*smax + s;
					//if (!cl_shadows[lnum].dark)
					//	blocklights[i] += temp;
		//			else
					{
						if (blocklights[i] > temp)
							blocklights[i] -= temp;
						else
							blocklights[i] = 0;
					}
				}




			}
		}
	}
}


void R_AddShadowsRGB ()
{
	msurface_t *surf;

	int		lnum;
	int		sd, td;
	float		dist, rad, minlight;
	vec3_t		impact, local;
	vec3_t      origin_for_ent; // mankrip - dynamic lights on moving brush models fix
	int		s, t;
	int		i;
	int		smax, tmax;
	mtexinfo_t	*tex;
	
// 2001-09-11 Colored lightning by LordHavoc/Sarcazm/Maddes  start
	float		cred, cgreen, cblue, brightness;
	unsigned	*bl;
// 2001-09-11 Colored lightning by LordHavoc/Sarcazm/Maddes  end
	surf = r_drawsurf.surf;
	smax = (surf->extents[0]>>4)+1;
	tmax = (surf->extents[1]>>4)+1;
	tex = surf->texinfo;

	for (lnum=0 ; lnum<MAX_SHADOWS ; lnum++)
	{
		if ( !(surf->shadowbits & (1<<lnum) ) )
			continue;		// not lit by this light

		rad = cl_shadows[lnum].radius;
		 VectorSubtract (cl_shadows[lnum].origin, currententity->origin, origin_for_ent); // mankrip - dynamic lights on moving brush models fix
		 	//	dist = DotProduct (cl_shadows[lnum].origin, surf->plane->normal) -
      dist = DotProduct (origin_for_ent, surf->plane->normal) - // mankrip - dynamic lights on moving brush models fix - edited
				surf->plane->dist;
		rad -= fabs(dist);
		minlight = cl_shadows[lnum].minlight;
		if (rad < minlight)
			continue;
		minlight = rad - minlight;

		for (i=0 ; i<3 ; i++)
		{
		 impact[i] = origin_for_ent[i] - // mankrip - dynamic lights on moving brush models fix - edited
			 //impact[i] = cl_shadows[lnum].origin[i] -
					surf->plane->normal[i]*dist;
		}

		local[0] = DotProduct (impact, tex->vecs[0]) + tex->vecs[0][3];
		local[1] = DotProduct (impact, tex->vecs[1]) + tex->vecs[1][3];

		local[0] -= surf->texturemins[0];
		local[1] -= surf->texturemins[1];

// 2001-09-11 Colored lightning by LordHavoc/Sarcazm/Maddes  start
		cred = -256;
		cgreen = -256;
		cblue = -256;

		bl = blocklights;
// 2001-09-11 Colored lightning by LordHavoc/Sarcazm/Maddes  end

		for (t = 0 ; t<tmax ; t++)
		{
			td = local[1] - t*16;
			if (td < 0)
				td = -td;
			for (s=0 ; s<smax ; s++)
			{
				sd = local[0] - s*16;
				if (sd < 0)
					sd = -sd;
				if (sd > td)
					dist = sd + (td>>1);
				else
					dist = td + (sd>>1);
				if (dist < minlight)
// 2001-09-11 Colored lightning by LordHavoc/Sarcazm/Maddes  start
//					blocklights[t*smax + s] += (rad - dist)*256;
				{
					brightness = rad - dist;
					bl[0] += (int) (brightness * cred);
					bl[1] += (int) (brightness * cgreen);
					bl[2] += (int) (brightness * cblue);
				}
				bl += 3;
// 2001-09-11 Colored lightning by LordHavoc/Sarcazm/Maddes  end
			}
		}
	}
}


/*
===============
R_BuildLightMap

Combine and scale multiple lightmaps into the 8.8 format in blocklights
===============
*/
void R_BuildLightMap (void)
{
	int			smax, tmax;
	int			t;
	int			i, size;
	byte		*lightmap;
	unsigned	scale;
	int			maps;
	msurface_t	*surf;

	surf = r_drawsurf.surf;

	smax = (surf->extents[0]>>4)+1;
	tmax = (surf->extents[1]>>4)+1;
	size = smax*tmax;
	lightmap = surf->samples;

	if (r_fullbright->value || !cl.worldmodel->lightdata)
	{
		for (i=0 ; i<size ; i++)
			blocklights[i] = 0;
		return;
	}

// clear to ambient
	for (i=0 ; i<size ; i++)
		blocklights[i] = r_refdef.ambientlight<<8;


// add all the lightmaps
	if (lightmap)
		for (maps = 0 ; maps < MAXLIGHTMAPS && surf->styles[maps] != 255 ;
			 maps++)
		{
			scale = r_drawsurf.lightadj[maps];	// 8.8 fraction		
			for (i=0 ; i<size ; i++)
				blocklights[i] += lightmap[i] * scale;
			lightmap += size;	// skip to next lightmap
		}

// add all the dynamic lights
			 if (surf->dlightframe == r_framecount){
		R_AddDynamicLights ();
			 }

// and shadows
			 if (surf->shadowframe == r_framecount){
				R_AddShadows ();
			 }			 


// bound, invert, and shift
	for (i=0 ; i<size ; i++)
	{
		t = (255*256 - (int)blocklights[i]) >> (8 - VID_CBITS);

		if (t < (1 << 6))
			t = (1 << 6);

		blocklights[i] = t;
	}
}




// fteqw version
void R_BuildLightMapRGB (void)
{
	int			smax, tmax;
	int			i, size;
	qbyte		*lightmap;
	unsigned	scale;
	int			maps;
	msurface_t	*surf;
	int r;
	surf = r_drawsurf.surf;

	smax = (surf->extents[0]>>4)+1;
	tmax = (surf->extents[1]>>4)+1;
	size = smax*tmax*3;
	lightmap = surf->samples;

	if (/* r_fullbright.value || */ !cl.worldmodel->lightdata)
	{
		for (i=0 ; i<size ; i++)
			blocklights[i] = 0;
		return;
	}

// clear to ambient
	for (i=0 ; i<size ; i++)
		blocklights[i] = r_refdef.ambientlight<<8;


// add all the lightmaps
	if (lightmap)
		for (maps = 0 ; maps < MAXLIGHTMAPS && surf->styles[maps] != 255 ;
			 maps++)
		{
			scale = r_drawsurf.lightadj[maps];	// 8.8 fraction		
			for (i=0 ; i<size ; i+=3)
			{
				blocklights[i]		+= lightmap[i] * scale;
				blocklights[i+1]	+= lightmap[i+1] * scale;
				blocklights[i+2]	+= lightmap[i+2] * scale;
			}
			lightmap += size;	// skip to next lightmap
		}

// add all the dynamic lights
			 if (surf->dlightframe == r_framecount)
		R_AddDynamicLightsRGB ();

// and shadows
			 if (surf->shadowframe == r_framecount)
				R_AddShadowsRGB ();
			 			 

	
// bound, invert, and shift
for (i=0 ; i<size ; i++)

	{


	
		r = blocklights[i];
		    blocklights[i] = (r < 0) ? 0 : (r > 65535) ? 65535 : r;
	}
}

/*
===============
R_TextureAnimation

Returns the proper texture for a given time and base texture
===============
*/
texture_t *R_TextureAnimation (texture_t *base)
{
	int		reletive;
	int		count;

	if (currententity->frame)
	{
		if (base->alternate_anims)
			base = base->alternate_anims;
	}

	if (!base->anim_total)
		return base;

	reletive = (int)(cl.time*10) % base->anim_total;

	count = 0;
	while (base->anim_min > reletive || base->anim_max <= reletive)
	{
		base = base->anim_next;
		if (!base)
			base = base;					// leilei - TFN bb2 workaround
//			Sys_Error ("R_TextureAnimation: broken cycle");	
		if (++count > 100)
			Sys_Error ("R_TextureAnimation: infinite cycle");
	}

	return base;
}

int	hqlite;
int	ditheredrend;		// dithering


extern	cvar_t *r_lightquality;
extern	cvar_t *r_dither;
extern byte	palmap2[64][64][64];		// Higher quality for lighting
/*
===============
R_DrawSurface
===============
*/
void R_DrawSurface (void)
{
	unsigned char	*basetptr;
	int				smax, tmax, twidth;
	int				u;

	int				soffset, basetoffset, texwidth;
	int				horzblockstep;
	unsigned char	*pcolumndest;
	void			(*pblockdrawer)(void);
	texture_t		*mt;
	
	if (!palmap2){
		hqlite = 0;	// don't do 18-bit lighting if we don't have this table
		coloredlights = 0; // disable colored lights anyway. :(
	}
	else
		hqlite = 1; // always use hqlite
	// calculate the lightings
	if (coloredlights)
	R_BuildLightMapRGB ();
	else
	R_BuildLightMap ();


// calculate the lightings
	
	surfrowbytes = r_drawsurf.rowbytes;

	mt = r_drawsurf.texture;

	r_source = (byte *)mt + mt->offsets[r_drawsurf.surfmip];


// the fractional light values should range from 0 to (VID_GRADES - 1) << 16
// from a source range of 0 - 255

	texwidth = mt->width >> r_drawsurf.surfmip;

	blocksize = 16 >> r_drawsurf.surfmip;
	blockdivshift = 4 - r_drawsurf.surfmip;
	blockdivmask = (1 << blockdivshift) - 1;

	r_lightwidth = (r_drawsurf.surf->extents[0]>>4)+1;

	r_numhblocks = r_drawsurf.surfwidth >> blockdivshift;
	r_numvblocks = r_drawsurf.surfheight >> blockdivshift;

//==============================
	{
		if (lowworld){
			if (coloredlights){
		//		if (r_dither->value)pblockdrawer = surfmiptable8RGBXD[r_drawsurf.surfmip]; // 18-bit lookups
		//		else
				if (ditheredrend)	pblockdrawer = surfmiptable8RGBXD[r_drawsurf.surfmip]; // 18-bit lookups
				else
					pblockdrawer = surfmiptable8RGBX[r_drawsurf.surfmip]; // 18-bit lookups
			}
		else
		pblockdrawer = surfmiptablefast[r_drawsurf.surfmip];	// use old fashioned grays
		}
		else
		{
			if (coloredlights){
					if (ditheredrend)		pblockdrawer = surfmiptable8RGBXD[r_drawsurf.surfmip]; // 18-bit lookups
								else
					pblockdrawer = surfmiptable8RGBX[r_drawsurf.surfmip]; // 18-bit lookups
			}
		else
		pblockdrawer = surfmiptable[r_drawsurf.surfmip];	// use old fashioned grays
		}
	// TODO: only needs to be set when there is a display settings change
		horzblockstep = blocksize;
	}
	smax = mt->width >> r_drawsurf.surfmip;
	twidth = texwidth;
	tmax = mt->height >> r_drawsurf.surfmip;
	sourcetstep = texwidth;
	r_stepback = tmax * twidth;

	r_sourcemax = r_source + (tmax * smax);

	soffset = r_drawsurf.surf->texturemins[0];
	basetoffset = r_drawsurf.surf->texturemins[1];

// << 16 components are to guarantee positive values for %
	soffset = ((soffset >> r_drawsurf.surfmip) + (smax << 16)) % smax;
	basetptr = &r_source[((((basetoffset >> r_drawsurf.surfmip)
		+ (tmax << 16)) % tmax) * twidth)];

	pcolumndest = r_drawsurf.surfdat;


	if (ditheredrend)
	horzblockstep *= 2;

	for (u=0 ; u<r_numhblocks; u++)
	{

	if (coloredlights)	
		r_lightptr = blocklights + u * 3;
	else
		r_lightptr = blocklights + u;

		prowdestbase = pcolumndest;

		pbasesource = basetptr + soffset;

		(*pblockdrawer)();

		soffset = soffset + blocksize;
		if (soffset >= smax)
			soffset = 0;

		pcolumndest += horzblockstep;
	}


	
}


//=============================================================================

#if	!id386



/*
================
R_DrawSurfaceBlock8_mip0
================
*/
void R_DrawSurfaceBlock8_mip0 (void)
{
	int				v, i, b, lightstep, lighttemp, light;
	unsigned char	pix, *psource, *prowdest;

	psource = pbasesource;
	prowdest = prowdestbase;

	for (v=0 ; v<r_numvblocks ; v++)
	{
	// FIXME: make these locals?
	// FIXME: use delta rather than both right and left, like ASM?
		lightleft = r_lightptr[0];
		lightright = r_lightptr[1];
		r_lightptr += r_lightwidth;
		
		lightleftstep = (r_lightptr[0] - lightleft) >> 4;
		
		lightrightstep = (r_lightptr[1] - lightright) >> 4;

		for (i=0 ; i<16 ; i++)
		{
			lighttemp = lightleft - lightright;
			lightstep = lighttemp >> 4;

			light = lightright;

			for (b=15; b>=0; b--)
			{
				pix = psource[b];
				prowdest[b] = ((unsigned char *)vid.colormap)
					[(light & 0xFF00) + pix];

			
				light += lightstep;
			}

			psource += sourcetstep;
				
			lightright += lightrightstep;
				
			lightleft += lightleftstep;
			prowdest += surfrowbytes;
		}

		if (psource >= r_sourcemax)
			psource -= r_stepback;
	}
}

/*
================
R_DrawSurfaceBlock8_mip1
================
*/
void R_DrawSurfaceBlock8_mip1 (void)
{
	int				v, i, b, lightstep, lighttemp, light;
	unsigned char	pix, *psource, *prowdest;

	psource = pbasesource;
	prowdest = prowdestbase;

	for (v=0 ; v<r_numvblocks ; v++)
	{
	// FIXME: make these locals?
	// FIXME: use delta rather than both right and left, like ASM?
		lightleft = r_lightptr[0];
		lightright = r_lightptr[1];
		r_lightptr += r_lightwidth;
		
		lightleftstep = (r_lightptr[0] - lightleft) >> 3;

		lightrightstep = (r_lightptr[1] - lightright) >> 3;

		for (i=0 ; i<8 ; i++)
		{
			lighttemp = lightleft - lightright;
			lightstep = lighttemp >> 3;

			light = lightright;

			for (b=7; b>=0; b--)
			{
				pix = psource[b];
				prowdest[b] = ((unsigned char *)vid.colormap)
						[(light & 0xFF00) + pix];
				
				light += lightstep;
			}

			psource += sourcetstep;
			
			lightright += lightrightstep;
			
			lightleft += lightleftstep;
			prowdest += surfrowbytes;
		}

		if (psource >= r_sourcemax)
			psource -= r_stepback;
	}
}


/*
================
R_DrawSurfaceBlock8_mip2
================
*/
void R_DrawSurfaceBlock8_mip2 (void)
{
	int				v, i, b, lightstep, lighttemp, light;
	unsigned char	pix, *psource, *prowdest;

	psource = pbasesource;
	prowdest = prowdestbase;

	for (v=0 ; v<r_numvblocks ; v++)
	{
	// FIXME: make these locals?
	// FIXME: use delta rather than both right and left, like ASM?
		lightleft = r_lightptr[0];
		lightright = r_lightptr[1];
		r_lightptr += r_lightwidth;
		
		lightleftstep = (r_lightptr[0] - lightleft) >> 2;
		
		lightrightstep = (r_lightptr[1] - lightright) >> 2;

		for (i=0 ; i<4 ; i++)
		{
			lighttemp = lightleft - lightright;
			lightstep = lighttemp >> 2;

			light = lightright;

			for (b=3; b>=0; b--)
			{
				pix = psource[b];
				prowdest[b] = ((unsigned char *)vid.colormap)
						[(light & 0xFF00) + pix];
				
				light += lightstep;
			}

			psource += sourcetstep;
			lightright += lightrightstep;
			lightleft += lightleftstep;
			prowdest += surfrowbytes;
		}

		if (psource >= r_sourcemax)
			psource -= r_stepback;
	}
}


/*
================
R_DrawSurfaceBlock8_mip3
================
*/
void R_DrawSurfaceBlock8_mip3 (void)
{
	int				v, i, b, lightstep, lighttemp, light;
	unsigned char	pix, *psource, *prowdest;

	psource = pbasesource;
	prowdest = prowdestbase;

	for (v=0 ; v<r_numvblocks ; v++)
	{
	// FIXME: make these locals?
	// FIXME: use delta rather than both right and left, like ASM?
		lightleft = r_lightptr[0];
		lightright = r_lightptr[1];
		r_lightptr += r_lightwidth;
		lightleftstep = (r_lightptr[0] - lightleft) >> 1;
		lightrightstep = (r_lightptr[1] - lightright) >> 1;

		for (i=0 ; i<2 ; i++)
		{
			lighttemp = lightleft - lightright;
			lightstep = lighttemp >> 1;

			light = lightright;

			for (b=1; b>=0; b--)
			{
				pix = psource[b];
				prowdest[b] = ((unsigned char *)vid.colormap)
						[(light & 0xFF00) + pix];
				light += lightstep;
			}

			psource += sourcetstep;
			lightright += lightrightstep;
			lightleft += lightleftstep;
			prowdest += surfrowbytes;
		}

		if (psource >= r_sourcemax)
			psource -= r_stepback;
	}
}


/*
================
R_DrawSurfaceBlock16

FIXME: make this work
================
*/
void R_DrawSurfaceBlock16 (void)
{
	int				k;
	unsigned char	*psource;
	int				lighttemp, lightstep, light;
	unsigned short	*prowdest;

	prowdest = (unsigned short *)prowdestbase;

	for (k=0 ; k<blocksize ; k++)
	{
		unsigned short	*pdest;
		unsigned char	pix;
		int				b;

		psource = pbasesource;
		lighttemp = lightright - lightleft;
		lightstep = lighttemp >> blockdivshift;

		light = lightleft;
		pdest = prowdest;

		for (b=0; b<blocksize; b++)
		{
			pix = *psource;
			*pdest = vid.colormap16[(light & 0xFF00) + pix];
			psource += sourcesstep;
			pdest++;
			light += lightstep;
		}

		pbasesource += sourcetstep;
		lightright += lightrightstep;
		lightleft += lightleftstep;
		prowdest = (unsigned short *)((long)prowdest + surfrowbytes);
	}

	prowdestbase = prowdest;
}








#endif

#if id386
// FAST VERSIONS BELOW LACK LIGHTSTEPPING


void R_DrawSurfaceBlock8_mip0fast (void)
{
	int				v, i, b, light;
	unsigned char	pix, *psource, *prowdest;

	psource = pbasesource;
	prowdest = prowdestbase;

	for (v=0 ; v<r_numvblocks ; v++)
	{
		light = r_lightptr[0];
		for (i=0 ; i<16 ; i++)
		{
			for (b=15; b>=0; b--)
			{
				pix = psource[b];
				prowdest[b] = ((unsigned char *)vid.colormap)
					[(light & 0xFF00) + pix];
			}
			prowdest += surfrowbytes;
		}

		if (psource >= r_sourcemax)
			psource -= r_stepback;
	}
}

void R_DrawSurfaceBlock8_mip1fast (void)
{
	int				v, i, b, light;
	unsigned char	pix, *psource, *prowdest;

	psource = pbasesource;
	prowdest = prowdestbase;

	for (v=0 ; v<r_numvblocks ; v++)
	{
		light = r_lightptr[0];
		for (i=0 ; i<8 ; i++)
		{
			for (b=7; b>=0; b--)
			{
				pix = psource[b];
				prowdest[b] = ((unsigned char *)vid.colormap)
					[(light & 0xFF00) + pix];
			}
			prowdest += surfrowbytes;
		}

		if (psource >= r_sourcemax)
			psource -= r_stepback;
	}
}

void R_DrawSurfaceBlock8_mip2fast (void)
{
	int				v, i, b, light;
	unsigned char	pix, *psource, *prowdest;

	psource = pbasesource;
	prowdest = prowdestbase;

	for (v=0 ; v<r_numvblocks ; v++)
	{
		light = r_lightptr[0];
		for (i=0 ; i<4 ; i++)
		{
			for (b=3; b>=0; b--)
			{
				pix = psource[b];
				prowdest[b] = ((unsigned char *)vid.colormap)
					[(light & 0xFF00) + pix];
			}
			prowdest += surfrowbytes;
		}

		if (psource >= r_sourcemax)
			psource -= r_stepback;
	}
}

void R_DrawSurfaceBlock8_mip3fast (void)
{
	int				v, i, b, light;
	unsigned char	pix, *psource, *prowdest;

	psource = pbasesource;
	prowdest = prowdestbase;

	for (v=0 ; v<r_numvblocks ; v++)
	{
		light = r_lightptr[0];
		for (i=0 ; i<2 ; i++)
		{
			for (b=1; b>=0; b--)
			{
				pix = psource[b];
				prowdest[b] = ((unsigned char *)vid.colormap)
					[(light & 0xFF00) + pix];
			}
			prowdest += surfrowbytes;
		}

		if (psource >= r_sourcemax)
			psource -= r_stepback;
	}
}



#endif







extern unsigned	d_8to24table[256];
extern byte		*lmmap;
extern byte	palmap[32][32][32];	
extern byte	palmap2[64][64][64];	





//	=!=!=!=!=!=!=!=
// ASMME: surf8rgb.s
//	=!=!=!=!=!=!=!=






// final!!! Crap cleaned out and slightly more optimized

extern byte transTable[256][256];


void R_DrawSurfaceBlock8RGBXDuffTest (int mip, int mif)
{
	int				v, i, b; 
	int		mil;
	int lightstep[3], lighttemp[3], light[3];
	int lightdelta[3], lightdeltastep[3];
	int		r;
	unsigned char	pix, *psource, *prowdest;
	register int count, n;
  
	unsigned char *pix24;
		int trans[3];
	psource = pbasesource;
	
	prowdest = prowdestbase;

	
	mil = mip - 1;	


	for (v=0 ; v<r_numvblocks ; v++)
	{
	// FIXME: make these locals?
	// FIXME: use delta rather than both right and left, like ASM?

		lightlefta[0] = r_lightptr[0];
		lightrighta[0] = r_lightptr[3];
		lightlefta[1] = r_lightptr[0+1];
		lightrighta[1] = r_lightptr[3+1];
		lightlefta[2] = r_lightptr[0+2];
		lightrighta[2] = r_lightptr[3+2];
		
		lightdelta[0] = (lightlefta[0] - lightrighta[0])  >> mif; 
		lightdelta[1] = (lightlefta[1] - lightrighta[1])  >> mif;  
		lightdelta[2] = (lightlefta[2] - lightrighta[2])  >> mif; 


		r_lightptr += r_lightwidth * 3;

		lightleftstepa[0] = (r_lightptr[0] - lightlefta[0]) >> mif;
		lightrightstepa[0] = (r_lightptr[3] - lightrighta[0]) >> mif;

		lightleftstepa[1] = (r_lightptr[0+1] - lightlefta[1]) >> mif;
		lightrightstepa[1] = (r_lightptr[3+1] - lightrighta[1]) >> mif;

		lightleftstepa[2] = (r_lightptr[0+2] - lightlefta[2]) >> mif;
		lightrightstepa[2] = (r_lightptr[3+2] - lightrighta[2]) >> mif;

		lightdeltastep[0] = (lightleftstepa[0] - lightrightstepa[0]) >> mif;
		lightdeltastep[1] = (lightleftstepa[1] - lightrightstepa[1]) >> mif;
		lightdeltastep[2] = (lightleftstepa[2] - lightrightstepa[2]) >> mif;

		for (i=0 ; i<mip ; i++)
		{

			light[0] = lightrighta[0];
			light[1] = lightrighta[1];
			light[2] = lightrighta[2];
			// leilei's ugly arse Duff device
			count = mil;
			n = (count + 15) / 16;
			switch(count % 15){
			case 0: do{
			case 15: if (psource[15] < host_fullbrights){	pix = psource[15]; pix24 = (unsigned char *)&d_8to24table[pix]; trans[0] = (pix24[0] * (light[0])) >> 17;	trans[1] = (pix24[1] * (light[1])) >> 17; trans[2] = (pix24[2] * (light[2])) >> 17;	if (trans[0] < 0) trans[0] = 0;	if (trans[1] < 0) trans[1] = 0;	if (trans[2] < 0) trans[2] = 0;	if (trans[0] > 63) trans[0] = 63; if (trans[1] > 63) trans[1] = 63;	if (trans[2] > 63) trans[2] = 63; prowdest[15] = palmap2[trans[0]][trans[1]][trans[2]];	} else prowdest[15] = psource[15]; light[0] += lightdelta[0]; light[1] += lightdelta[1]; light[2] += lightdelta[2];
			case 14: if (psource[14] < host_fullbrights){	pix = psource[14]; pix24 = (unsigned char *)&d_8to24table[pix]; trans[0] = (pix24[0] * (light[0])) >> 17;	trans[1] = (pix24[1] * (light[1])) >> 17; trans[2] = (pix24[2] * (light[2])) >> 17;	if (trans[0] < 0) trans[0] = 0;	if (trans[1] < 0) trans[1] = 0;	if (trans[2] < 0) trans[2] = 0;	if (trans[0] > 63) trans[0] = 63; if (trans[1] > 63) trans[1] = 63;	if (trans[2] > 63) trans[2] = 63; prowdest[14] = palmap2[trans[0]][trans[1]][trans[2]];	} else prowdest[14] = psource[14]; light[0] += lightdelta[0]; light[1] += lightdelta[1]; light[2] += lightdelta[2];
			case 13: if (psource[13] < host_fullbrights){	pix = psource[13]; pix24 = (unsigned char *)&d_8to24table[pix]; trans[0] = (pix24[0] * (light[0])) >> 17;	trans[1] = (pix24[1] * (light[1])) >> 17; trans[2] = (pix24[2] * (light[2])) >> 17;	if (trans[0] < 0) trans[0] = 0;	if (trans[1] < 0) trans[1] = 0;	if (trans[2] < 0) trans[2] = 0;	if (trans[0] > 63) trans[0] = 63; if (trans[1] > 63) trans[1] = 63;	if (trans[2] > 63) trans[2] = 63; prowdest[13] = palmap2[trans[0]][trans[1]][trans[2]];	} else prowdest[13] = psource[13]; light[0] += lightdelta[0]; light[1] += lightdelta[1]; light[2] += lightdelta[2];
			case 12: if (psource[12] < host_fullbrights){	pix = psource[12]; pix24 = (unsigned char *)&d_8to24table[pix]; trans[0] = (pix24[0] * (light[0])) >> 17;	trans[1] = (pix24[1] * (light[1])) >> 17; trans[2] = (pix24[2] * (light[2])) >> 17;	if (trans[0] < 0) trans[0] = 0;	if (trans[1] < 0) trans[1] = 0;	if (trans[2] < 0) trans[2] = 0;	if (trans[0] > 63) trans[0] = 63; if (trans[1] > 63) trans[1] = 63;	if (trans[2] > 63) trans[2] = 63; prowdest[12] = palmap2[trans[0]][trans[1]][trans[2]];	} else prowdest[12] = psource[12]; light[0] += lightdelta[0]; light[1] += lightdelta[1]; light[2] += lightdelta[2];
			case 11: if (psource[11] < host_fullbrights){	pix = psource[11]; pix24 = (unsigned char *)&d_8to24table[pix]; trans[0] = (pix24[0] * (light[0])) >> 17;	trans[1] = (pix24[1] * (light[1])) >> 17; trans[2] = (pix24[2] * (light[2])) >> 17;	if (trans[0] < 0) trans[0] = 0;	if (trans[1] < 0) trans[1] = 0;	if (trans[2] < 0) trans[2] = 0;	if (trans[0] > 63) trans[0] = 63; if (trans[1] > 63) trans[1] = 63;	if (trans[2] > 63) trans[2] = 63; prowdest[11] = palmap2[trans[0]][trans[1]][trans[2]];	} else prowdest[11] = psource[11]; light[0] += lightdelta[0]; light[1] += lightdelta[1]; light[2] += lightdelta[2];
			case 10: if (psource[10] < host_fullbrights){	pix = psource[10]; pix24 = (unsigned char *)&d_8to24table[pix]; trans[0] = (pix24[0] * (light[0])) >> 17;	trans[1] = (pix24[1] * (light[1])) >> 17; trans[2] = (pix24[2] * (light[2])) >> 17;	if (trans[0] < 0) trans[0] = 0;	if (trans[1] < 0) trans[1] = 0;	if (trans[2] < 0) trans[2] = 0;	if (trans[0] > 63) trans[0] = 63; if (trans[1] > 63) trans[1] = 63;	if (trans[2] > 63) trans[2] = 63; prowdest[10] = palmap2[trans[0]][trans[1]][trans[2]];	} else prowdest[10] = psource[10]; light[0] += lightdelta[0]; light[1] += lightdelta[1]; light[2] += lightdelta[2];
			case 9: if (psource[9] < host_fullbrights){	pix = psource[9]; pix24 = (unsigned char *)&d_8to24table[pix]; trans[0] = (pix24[0] * (light[0])) >> 17;	trans[1] = (pix24[1] * (light[1])) >> 17; trans[2] = (pix24[2] * (light[2])) >> 17;	if (trans[0] < 0) trans[0] = 0;	if (trans[1] < 0) trans[1] = 0;	if (trans[2] < 0) trans[2] = 0;	if (trans[0] > 63) trans[0] = 63; if (trans[1] > 63) trans[1] = 63;	if (trans[2] > 63) trans[2] = 63; prowdest[9] = palmap2[trans[0]][trans[1]][trans[2]];	} else prowdest[9] = psource[9]; light[0] += lightdelta[0]; light[1] += lightdelta[1]; light[2] += lightdelta[2];
			case 8: if (psource[8] < host_fullbrights){	pix = psource[8]; pix24 = (unsigned char *)&d_8to24table[pix]; trans[0] = (pix24[0] * (light[0])) >> 17;	trans[1] = (pix24[1] * (light[1])) >> 17; trans[2] = (pix24[2] * (light[2])) >> 17;	if (trans[0] < 0) trans[0] = 0;	if (trans[1] < 0) trans[1] = 0;	if (trans[2] < 0) trans[2] = 0;	if (trans[0] > 63) trans[0] = 63; if (trans[1] > 63) trans[1] = 63;	if (trans[2] > 63) trans[2] = 63; prowdest[8] = palmap2[trans[0]][trans[1]][trans[2]];	} else prowdest[8] = psource[8]; light[0] += lightdelta[0]; light[1] += lightdelta[1]; light[2] += lightdelta[2];
			case 7: if (psource[7] < host_fullbrights){	pix = psource[7]; pix24 = (unsigned char *)&d_8to24table[pix]; trans[0] = (pix24[0] * (light[0])) >> 17;	trans[1] = (pix24[1] * (light[1])) >> 17; trans[2] = (pix24[2] * (light[2])) >> 17;	if (trans[0] < 0) trans[0] = 0;	if (trans[1] < 0) trans[1] = 0;	if (trans[2] < 0) trans[2] = 0;	if (trans[0] > 63) trans[0] = 63; if (trans[1] > 63) trans[1] = 63;	if (trans[2] > 63) trans[2] = 63; prowdest[7] = palmap2[trans[0]][trans[1]][trans[2]];	} else prowdest[7] = psource[7]; light[0] += lightdelta[0]; light[1] += lightdelta[1]; light[2] += lightdelta[2];
			case 6: if (psource[6] < host_fullbrights){	pix = psource[6]; pix24 = (unsigned char *)&d_8to24table[pix]; trans[0] = (pix24[0] * (light[0])) >> 17;	trans[1] = (pix24[1] * (light[1])) >> 17; trans[2] = (pix24[2] * (light[2])) >> 17;	if (trans[0] < 0) trans[0] = 0;	if (trans[1] < 0) trans[1] = 0;	if (trans[2] < 0) trans[2] = 0;	if (trans[0] > 63) trans[0] = 63; if (trans[1] > 63) trans[1] = 63;	if (trans[2] > 63) trans[2] = 63; prowdest[6] = palmap2[trans[0]][trans[1]][trans[2]];	} else prowdest[6] = psource[6]; light[0] += lightdelta[0]; light[1] += lightdelta[1]; light[2] += lightdelta[2];
			case 5: if (psource[5] < host_fullbrights){	pix = psource[5]; pix24 = (unsigned char *)&d_8to24table[pix]; trans[0] = (pix24[0] * (light[0])) >> 17;	trans[1] = (pix24[1] * (light[1])) >> 17; trans[2] = (pix24[2] * (light[2])) >> 17;	if (trans[0] < 0) trans[0] = 0;	if (trans[1] < 0) trans[1] = 0;	if (trans[2] < 0) trans[2] = 0;	if (trans[0] > 63) trans[0] = 63; if (trans[1] > 63) trans[1] = 63;	if (trans[2] > 63) trans[2] = 63; prowdest[5] = palmap2[trans[0]][trans[1]][trans[2]];	} else prowdest[5] = psource[5]; light[0] += lightdelta[0]; light[1] += lightdelta[1]; light[2] += lightdelta[2];
			case 4: if (psource[4] < host_fullbrights){	pix = psource[4]; pix24 = (unsigned char *)&d_8to24table[pix]; trans[0] = (pix24[0] * (light[0])) >> 17;	trans[1] = (pix24[1] * (light[1])) >> 17; trans[2] = (pix24[2] * (light[2])) >> 17;	if (trans[0] < 0) trans[0] = 0;	if (trans[1] < 0) trans[1] = 0;	if (trans[2] < 0) trans[2] = 0;	if (trans[0] > 63) trans[0] = 63; if (trans[1] > 63) trans[1] = 63;	if (trans[2] > 63) trans[2] = 63; prowdest[4] = palmap2[trans[0]][trans[1]][trans[2]];	} else prowdest[4] = psource[4]; light[0] += lightdelta[0]; light[1] += lightdelta[1]; light[2] += lightdelta[2];
			case 3: if (psource[3] < host_fullbrights){	pix = psource[3]; pix24 = (unsigned char *)&d_8to24table[pix]; trans[0] = (pix24[0] * (light[0])) >> 17;	trans[1] = (pix24[1] * (light[1])) >> 17; trans[2] = (pix24[2] * (light[2])) >> 17;	if (trans[0] < 0) trans[0] = 0;	if (trans[1] < 0) trans[1] = 0;	if (trans[2] < 0) trans[2] = 0;	if (trans[0] > 63) trans[0] = 63; if (trans[1] > 63) trans[1] = 63;	if (trans[2] > 63) trans[2] = 63; prowdest[3] = palmap2[trans[0]][trans[1]][trans[2]];	} else prowdest[3] = psource[3]; light[0] += lightdelta[0]; light[1] += lightdelta[1]; light[2] += lightdelta[2];
			case 2: if (psource[2] < host_fullbrights){	pix = psource[2]; pix24 = (unsigned char *)&d_8to24table[pix]; trans[0] = (pix24[0] * (light[0])) >> 17;	trans[1] = (pix24[1] * (light[1])) >> 17; trans[2] = (pix24[2] * (light[2])) >> 17;	if (trans[0] < 0) trans[0] = 0;	if (trans[1] < 0) trans[1] = 0;	if (trans[2] < 0) trans[2] = 0;	if (trans[0] > 63) trans[0] = 63; if (trans[1] > 63) trans[1] = 63;	if (trans[2] > 63) trans[2] = 63; prowdest[2] = palmap2[trans[0]][trans[1]][trans[2]];	} else prowdest[2] = psource[2]; light[0] += lightdelta[0]; light[1] += lightdelta[1]; light[2] += lightdelta[2];
			case 1: if (psource[1] < host_fullbrights){	pix = psource[1]; pix24 = (unsigned char *)&d_8to24table[pix]; trans[0] = (pix24[0] * (light[0])) >> 17;	trans[1] = (pix24[1] * (light[1])) >> 17; trans[2] = (pix24[2] * (light[2])) >> 17;	if (trans[0] < 0) trans[0] = 0;	if (trans[1] < 0) trans[1] = 0;	if (trans[2] < 0) trans[2] = 0;	if (trans[0] > 63) trans[0] = 63; if (trans[1] > 63) trans[1] = 63;	if (trans[2] > 63) trans[2] = 63; prowdest[1] = palmap2[trans[0]][trans[1]][trans[2]];	} else prowdest[1] = psource[1]; light[0] += lightdelta[0]; light[1] += lightdelta[1]; light[2] += lightdelta[2];
					}while (--n > 0);
			
		//	case 0: if (psource[0] < host_fullbrights){	pix = psource[0]; pix24 = (unsigned char *)&d_8to24table[pix]; trans[0] = (pix24[0] * (light[0])) >> 17;	trans[1] = (pix24[1] * (light[1])) >> 17; trans[2] = (pix24[2] * (light[2])) >> 17;	if (trans[0] < 0) trans[0] = 0;	if (trans[1] < 0) trans[1] = 0;	if (trans[2] < 0) trans[2] = 0;	if (trans[0] > 63) trans[0] = 63; if (trans[1] > 63) trans[1] = 63;	if (trans[2] > 63) trans[2] = 63; prowdest[0] = palmap2[trans[0]][trans[1]][trans[2]];	} else prowdest[0] = psource[0]; light[0] += lightdelta[0]; light[1] += lightdelta[1]; light[2] += lightdelta[2];

			}
	//		for (b=mil; b>=0; b--)
	//		{
					//	if (psource[b] < host_fullbrights){	pix = psource[b] pix24 = (unsigned char *)&d_8to24table[pix]; trans[0] = (pix24[0] * (light[0])) >> 17;	trans[1] = (pix24[1] * (light[1])) >> 17; trans[2] = (pix24[2] * (light[2])) >> 17;	if (trans[0] < 0) trans[0] = 0;	if (trans[1] < 0) trans[1] = 0;	if (trans[2] < 0) trans[2] = 0;	if (trans[0] > 63) trans[0] = 63; if (trans[1] > 63) trans[1] = 63;	if (trans[2] > 63) trans[2] = 63; prowdest[b] = palmap2[trans[0]][trans[1]][trans[2]];	} else prowdest[b] = psource[b]; light[0] += lightdelta[0]; light[1] += lightdelta[1]; light[2] += lightdelta[2];
			
			/*
						if (psource[b] < host_fullbrights){ 	// Preserve the fullbrights

						pix = psource[b];//
						pix24 = (unsigned char *)&d_8to24table[pix];
						// sadly, using a for loop makes this function much slower
							trans[0] = (pix24[0] * (light[0])) >> 17;
							trans[1] = (pix24[1] * (light[1])) >> 17;
							trans[2] = (pix24[2] * (light[2])) >> 17;
							if (trans[0] < 0) trans[0] = 0;	if (trans[1] < 0) trans[1] = 0;	if (trans[2] < 0) trans[2] = 0;
							if (trans[0] > 63) trans[0] = 63; if (trans[1] > 63) trans[1] = 63;	if (trans[2] > 63) trans[2] = 63;


						prowdest[b] = palmap2[trans[0]][trans[1]][trans[2]];
						
						}
						else
						prowdest[b] = psource[b];				// fullbright pixel

						light[0] += lightdelta[0];
						light[1] += lightdelta[1];
						light[2] += lightdelta[2];*/

	//		}
		
			psource += sourcetstep;
				for (r=0; r<3; r++){
					lightrighta[r] += lightrightstepa[r];
					lightlefta[r] += lightleftstepa[r];
					lightdelta[r] += lightdeltastep[r];
					}

			prowdest += surfrowbytes;
		}

		if (psource >= r_sourcemax)
			psource -= r_stepback;
		
	}
}



// leilei - i put them all here, cutting lots of bloat. I also deprecated hqlite. :(
void R_DrawSurfaceBlock8RGBX_mip0 (void)
{
	R_DrawSurfaceBlock8RGBX (16, 4);
}
void R_DrawSurfaceBlock8RGBX_mip1 (void)
{
	R_DrawSurfaceBlock8RGBX (8, 3);
}
void R_DrawSurfaceBlock8RGBX_mip2 (void)
{
	R_DrawSurfaceBlock8RGBX (4, 2);
}
void R_DrawSurfaceBlock8RGBX_mip3 (void)
{
	R_DrawSurfaceBlock8RGBX (2, 1);
}


// 18-bit version
// I tried to merge all the mips together and use deltas instead

void R_DrawSurfaceBlock8RGBX (int mip, int mif)
{
	int				v, i, b; 
	int		mil;
	int lightstep[3], lighttemp[3], light[3];
	int lightdelta[3], lightdeltastep[3];
	int		r;
	unsigned char	pix, *psource, *prowdest;

  
	unsigned char *pix24;
		int trans[3];
	psource = pbasesource;
	
	prowdest = prowdestbase;

	
	mil = mip - 1;	


	for (v=0 ; v<r_numvblocks ; v++)
	{
	// FIXME: make these locals?

		lightlefta[0] = r_lightptr[0];
		lightrighta[0] = r_lightptr[3];
		lightlefta[1] = r_lightptr[0+1];
		lightrighta[1] = r_lightptr[3+1];
		lightlefta[2] = r_lightptr[0+2];
		lightrighta[2] = r_lightptr[3+2];
		
		lightdelta[0] = (lightlefta[0] - lightrighta[0])  >> mif; 
		lightdelta[1] = (lightlefta[1] - lightrighta[1])  >> mif;  
		lightdelta[2] = (lightlefta[2] - lightrighta[2])  >> mif; 


		r_lightptr += r_lightwidth * 3;

		lightleftstepa[0] = (r_lightptr[0] - lightlefta[0]) >> mif;
		lightrightstepa[0] = (r_lightptr[3] - lightrighta[0]) >> mif;

		lightleftstepa[1] = (r_lightptr[0+1] - lightlefta[1]) >> mif;
		lightrightstepa[1] = (r_lightptr[3+1] - lightrighta[1]) >> mif;

		lightleftstepa[2] = (r_lightptr[0+2] - lightlefta[2]) >> mif;
		lightrightstepa[2] = (r_lightptr[3+2] - lightrighta[2]) >> mif;

		lightdeltastep[0] = (lightleftstepa[0] - lightrightstepa[0]) >> mif;
		lightdeltastep[1] = (lightleftstepa[1] - lightrightstepa[1]) >> mif;
		lightdeltastep[2] = (lightleftstepa[2] - lightrightstepa[2]) >> mif;

		for (i=0 ; i<mip ; i++)
		{
			light[0] = lightrighta[0];
			light[1] = lightrighta[1];
			light[2] = lightrighta[2];

		for (b=mil; b>=0; b--)
			{
	
						if (psource[b] < host_fullbrights){ 	// Preserve the fullbrights

						pix = psource[b];//
						pix24 = (unsigned char *)&d_8to24table[pix];
						// sadly, using a for loop makes this function much slower
							trans[0] = (pix24[0] * (light[0])) >> 17;
							trans[1] = (pix24[1] * (light[1])) >> 17;
							trans[2] = (pix24[2] * (light[2])) >> 17;

							if (trans[0] < 0) trans[0] = 0;	if (trans[1] < 0) trans[1] = 0;	if (trans[2] < 0) trans[2] = 0;
							if (trans[0] > 63) trans[0] = 63; if (trans[1] > 63) trans[1] = 63;	if (trans[2] > 63) trans[2] = 63;


						prowdest[b] = palmap2[trans[0]][trans[1]][trans[2]];
						
						}
						else
						prowdest[b] = psource[b];				// fullbright pixel

						light[0] += lightdelta[0];
						light[1] += lightdelta[1];
						light[2] += lightdelta[2];

			}
		
			psource += sourcetstep;
				for (r=0; r<3; r++){
					lightrighta[r] += lightrightstepa[r];
					lightlefta[r] += lightleftstepa[r];
					lightdelta[r] += lightdeltastep[r];
					}

			prowdest += surfrowbytes;
		}

		if (psource >= r_sourcemax)
			psource -= r_stepback;
		
	}
}






// Dithered version (andrewj)


void R_DrawSurfaceBlock8RGBXD (int mip, int mif)
{
	int				v, i, b; 
	int lightstep[3], lighttemp[3], light[3];
	unsigned char	pix, *psource;
  unsigned short *prowdest;
	int mil;
	int sh1, sh2;
	qbyte	pax, *pasource;
    int me[3];
	unsigned char *pix24;
		int trans[3];
	byte	ar, ge, be;
	mil = mip - 1;
	psource = pbasesource;
	pasource = pbasesource;
	prowdest = (unsigned short *)prowdestbase;


	for (v=0 ; v<r_numvblocks ; v++)
	{
	// FIXME: make these locals?
	// FIXME: use delta rather than both right and left, like ASM?
	
		lightlefta[0]  = r_lightptr[0];
		lightlefta[1]  = r_lightptr[0+1];
		lightlefta[2]  = r_lightptr[0+2];

		lightrighta[0] = r_lightptr[3];
		lightrighta[1] = r_lightptr[3+1];
		lightrighta[2] = r_lightptr[3+2];
		
		r_lightptr += r_lightwidth * 3;

		lightleftstepa[0]  = ((int)r_lightptr[0]   - lightlefta[0]) >> mif;
		lightleftstepa[1]  = ((int)r_lightptr[0+1] - lightlefta[1]) >> mif;
		lightleftstepa[2]  = ((int)r_lightptr[0+2] - lightlefta[2]) >> mif;

		lightrightstepa[0] = ((int)r_lightptr[3]   - lightrighta[0]) >> mif;
		lightrightstepa[1] = ((int)r_lightptr[3+1] - lightrighta[1]) >> mif;
		lightrightstepa[2] = ((int)r_lightptr[3+2] - lightrighta[2]) >> mif;


		for (i=0 ; i<mip ; i++)
		{
			lightstep[0] = (lightlefta[0] - lightrighta[0]) >> mif;
			lightstep[1] = (lightlefta[1] - lightrighta[1]) >> mif;
			lightstep[2] = (lightlefta[2] - lightrighta[2]) >> mif;

			light[0] = lightrighta[0];
			light[1] = lightrighta[1];
			light[2] = lightrighta[2];

		for (b=mil; b>=0; b--)
		{
	
			if (psource[b] < host_fullbrights) 	// Preserve the fullbrights
      {
			pix = psource[b];//
			pix24 = (unsigned char *)&d_8to24table[pix];

		trans[0] = ((int)pix24[0] * light[0]) >> 18;
		trans[1] = ((int)pix24[1] * light[1]) >> 18;
		trans[2] = ((int)pix24[2] * light[2]) >> 18;

		if (trans[0] < 0) trans[0] = 0;
		if (trans[1] < 0) trans[1] = 0;
		if (trans[2] < 0) trans[2] = 0;

		if (trans[0] > 31) trans[0] = 31;
		if (trans[1] > 31) trans[1] = 31;
		if (trans[2] > 31) trans[2] = 31;
			
			prowdest[b] = (trans[0] << 10) | (trans[1] << 5) | trans[2];

			}
			else{
			pix = psource[b];
			pix24 = (unsigned char *)&d_8to24table[pix];
			trans[0] = ((int)pix24[0]) >> 3;
			trans[1] = ((int)pix24[1]) >> 3;
			trans[2] = ((int)pix24[2]) >> 3;
			prowdest[b] =  (trans[0] << 10) | (trans[1] << 5) | trans[2]; // everybody loves redmond

			}

			light[0] += lightstep[0];
			light[1] += lightstep[1];
			light[2] += lightstep[2];
			}
		
			psource += sourcetstep;

			lightrighta[0] += lightrightstepa[0];
			lightrighta[1] += lightrightstepa[1];
			lightrighta[2] += lightrightstepa[2];

			lightlefta[0] += lightleftstepa[0];
			lightlefta[1] += lightleftstepa[1];
			lightlefta[2] += lightleftstepa[2];

			prowdest += surfrowbytes;
		}

		if (psource >= r_sourcemax)
			psource -= r_stepback;
		
	}
}






void R_DrawSurfaceBlock8RGBXD_mip0 (void)
{
	R_DrawSurfaceBlock8RGBXD (16, 4);
}
void R_DrawSurfaceBlock8RGBXD_mip1 (void)
{
	R_DrawSurfaceBlock8RGBXD (8, 3);
}
void R_DrawSurfaceBlock8RGBXD_mip2 (void)
{
	R_DrawSurfaceBlock8RGBXD (4, 2);
}
void R_DrawSurfaceBlock8RGBXD_mip3 (void)
{
	R_DrawSurfaceBlock8RGBXD (2, 1);
}



//============================================================================

/*
================
R_GenTurbTile
================

 



*/	  
		
		  
			
			  


				
				  
					
					  
						
						  
					
void R_GenTurbTile (pixel_t *pbasetex, void *pdest)
{
	int		*turb;
	int		i, j, s, t;
	byte	*pd;

	turb = sintable + ((int)(cl.time*SPEED)&(CYCLE-1));
	pd = (byte *)pdest;

	for (i=0 ; i<TILE_SIZE ; i++)
	{
		for (j=0 ; j<TILE_SIZE ; j++)
		{
			s = (((j << 16) + turb[i & (CYCLE-1)]) >> 16) & 63;
			t = (((i << 16) + turb[j & (CYCLE-1)]) >> 16) & 63;
			*pd++ = *(pbasetex + (t<<6) + s);
		}
	}
}


/*
================
R_GenTurbTile16
================
*/
void R_GenTurbTile16 (pixel_t *pbasetex, void *pdest)
{
	int				*turb;
	int				i, j, s, t;
	unsigned short	*pd;

	turb = sintable + ((int)(cl.time*SPEED)&(CYCLE-1));
	pd = (unsigned short *)pdest;

	for (i=0 ; i<TILE_SIZE ; i++)
	{
		for (j=0 ; j<TILE_SIZE ; j++)
		{
			s = (((j << 16) + turb[i & (CYCLE-1)]) >> 16) & 63;
			t = (((i << 16) + turb[j & (CYCLE-1)]) >> 16) & 63;
			*pd++ = d_8to16table[*(pbasetex + (t<<6) + s)];
		}
	}
}


/*
================
R_GenTile
================
*/
void R_GenTile (msurface_t *psurf, void *pdest)
{
	if (psurf->flags & SURF_DRAWTURB)
	{
		if (r_pixbytes == 1)
		{
			R_GenTurbTile ((pixel_t *)
				((byte *)psurf->texinfo->texture + psurf->texinfo->texture->offsets[0]), pdest);
		}
		else
		{
			R_GenTurbTile16 ((pixel_t *)
				((byte *)psurf->texinfo->texture + psurf->texinfo->texture->offsets[0]), pdest);
		}
	}
	else if (psurf->flags & SURF_DRAWSKY)
	{
		if (r_pixbytes == 1)
		{
			R_GenSkyTile (pdest);
		}
		else
		{
			R_GenSkyTile16 (pdest);
		}
	}
	else
	{
		Sys_Error ("Unknown tile type");
	}
}


















